import json
import os
import tempfile

import pytest

from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import ipuworkflowconfig
from leapp.libraries.stdlib import CalledProcessError
from leapp.models import OSRelease

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def _clean_leapp_envs(monkeypatch):
    """
    Clean all LEAPP environment variables before running the test to have
    fresh env.
    """
    for k, _ in os.environ.items():
        if k.startswith('LEAPP'):
            monkeypatch.delenv(k)


def _raise_call_error(*args):
    raise CalledProcessError(
        message='A Leapp Command Error occurred.',
        command=args,
        result={'signal': None, 'exit_code': 1, 'pid': 0, 'stdout': 'fake', 'stderr': 'fake'}
    )


def _get_os_release(version='7.9', codename='Maipo'):
    release = OSRelease(
        release_id='rhel',
        name='Red Hat Enterprise Linux Server',
        pretty_name='Red Hat Enterprise Linux',
        version='{} ({})'.format(version, codename),
        version_id='{}'.format(version),
        variant='Server',
        variant_id='server')
    return release


def test_leapp_env_vars(monkeypatch):
    _clean_leapp_envs(monkeypatch)
    monkeypatch.setenv('LEAPP_WHATEVER', '0')
    monkeypatch.setenv('LEAPP_VERBOSE', '1')
    monkeypatch.setenv('LEAPP_DEBUG', '1')
    monkeypatch.setenv('LEAPP_CURRENT_PHASE', 'test')
    monkeypatch.setenv('LEAPP_CURRENT_ACTOR', 'test')
    monkeypatch.setenv('TEST', 'test')
    monkeypatch.setenv('TEST2', 'test')

    assert len(ipuworkflowconfig.get_env_vars()) == 1


def test_get_os_release_info(monkeypatch):
    expected = _get_os_release('7.6')
    assert expected == ipuworkflowconfig.get_os_release(os.path.join(CUR_DIR, 'files/os-release'))

    with pytest.raises(StopActorExecutionError):
        ipuworkflowconfig.get_os_release(os.path.join(CUR_DIR, 'files/non-existent-file'))


def test_get_booted_kernel(monkeypatch):
    monkeypatch.setattr(ipuworkflowconfig, 'run', lambda x: {'stdout': '4.14.0-100.8.2.el7a.x86_64\n'})
    assert ipuworkflowconfig.get_booted_kernel() == '4.14.0-100.8.2.el7a.x86_64'

    monkeypatch.setattr(ipuworkflowconfig, 'run', _raise_call_error)
    with pytest.raises(StopActorExecutionError):
        ipuworkflowconfig.get_booted_kernel()



from collections import defaultdict, namedtuple
from functools import partial

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor import peseventsscanner_repomap
from leapp.libraries.actor.pes_event_parsing import Action, get_pes_events, Package
from leapp.libraries.common import rpms
from leapp.libraries.common.config import version
from leapp.libraries.stdlib import api
from leapp.libraries.stdlib.config import is_verbose
from leapp.models import (
    DistributionSignedRPM,
    EnabledModules,
    Module,
    PESIDRepositoryEntry,
    PESRpmTransactionTasks,
    RepositoriesBlacklisted,
    RepositoriesFacts,
    RepositoriesMapping,
    RepositoriesSetupTasks,
    RHUIInfo,
    RpmTransactionTasks
)

SKIPPED_PKGS_MSG = (
    'packages will be skipped because they are available only in '
    'target system repositories that are intentionally excluded from the '
    'list of repositories used during the upgrade. '
    'See the report message titled "Excluded target system repositories" '
    'for details.\nThe list of these packages:'
)


TransactionConfiguration = namedtuple('TransactionConfiguration', ('to_install', 'to_remove', 'to_keep'))


def get_cloud_provider_name(cloud_provider_variant):
    for cloud_provider_prefix in ('aws', 'azure', 'google'):
        if cloud_provider_variant.startswith(cloud_provider_prefix):
            return cloud_provider_prefix
    return cloud_provider_variant


def get_best_pesid_candidate(candidate_a, candidate_b, cloud_provider):
    cdn_candidate = None
    for candidate in (candidate_a, candidate_b):
        if candidate.rhui == cloud_provider:
            return candidate
        if not candidate.rhui:
            cdn_candidate = candidate

    # None of the candidate matches cloud provider and none of them is from CDN -
    # do not return anything as we don't want to get content from different cloud providers
    return cdn_candidate


def get_installed_pkgs():
    installed_pkgs = set()

    installed_rh_signed_rpm_msgs = api.consume(DistributionSignedRPM)
    installed_rh_signed_rpm_msg = next(installed_rh_signed_rpm_msgs, None)
    if list(installed_rh_signed_rpm_msgs):
        api.current_logger().warning('Unexpectedly received more than one DistributionSignedRPM message.')
    if not installed_rh_signed_rpm_msg:
        raise StopActorExecutionError('Cannot parse PES data properly due to missing list of installed packages',
                                      details={'Problem': 'Did not receive a message with installed Red Hat-signed '
                                                          'packages (DistributionSignedRPM)'})

    for pkg in installed_rh_signed_rpm_msg.items:
        modulestream = None
        if pkg.module and pkg.stream:
            modulestream = (pkg.module, pkg.stream)
        installed_pkgs.add(Package(name=pkg.name, repository=pkg.repository, modulestream=modulestream))

    return installed_pkgs


def get_transaction_configuration():
    """
    Get pkgs to install, keep and remove from the user configuration files in /etc/leapp/transaction/.

    These configuration files have higher priority than PES data.
    :return: RpmTransactionTasks model instance
    """
    transaction_configuration = TransactionConfiguration(to_install=[], to_remove=[], to_keep=[])

    _Pkg = partial(Package, repository=None, modulestream=None)

    for tasks in api.consume(RpmTransactionTasks):
        transaction_configuration.to_install.extend(_Pkg(name=pkg_name) for pkg_name in tasks.to_install)
        transaction_configuration.to_remove.extend(_Pkg(name=pkg_name) for pkg_name in tasks.to_remove)
        transaction_configuration.to_keep.extend(_Pkg(name=pkg_name) for pkg_name in tasks.to_keep)
    return transaction_configuration


def get_relevant_releases(events):
    """
    Get releases present in the PES Events that are relevant for this IPU.

    Relevant release happened between the source OS version and the target OS version.
    """
    # Collect releases that happened between source OS version and target OS version
    relevant_releases_match_list = [
        '> {0}'.format(api.current_actor().configuration.version.source),
        '<= {0}'.format(api.current_actor().configuration.version.target)
    ]
    releases = {event.to_release for event in events}
    releases = [r for r in releases if version.matches_version(relevant_releases_match_list, '{}.{}'.format(*r))]
    return sorted(releases)


def _get_enabled_modules():
    enabled_modules_msgs = api.consume(EnabledModules)
    enabled_modules_msg = next(enabled_modules_msgs, None)
    if list(enabled_modules_msgs):
        api.current_logger().warning('Unexpectedly received more than one EnabledModules message.')
    if not enabled_modules_msg:
        raise StopActorExecutionError('Cannot parse PES data properly due to missing list of enabled modules',
                                      details={'Problem': 'Did not receive a message with enabled module '
                                                          'streams (EnabledModules)'})
    return enabled_modules_msg.modules


def compute_pkg_changes_between_consequent_releases(source_installed_pkgs,
                                                    events,
                                                    release,
                                                    seen_pkgs,
                                                    pkgs_to_demodularize):
    logger = api.current_logger()
    # Start with the installed packages and modify the set according to release events
    target_pkgs = set(source_installed_pkgs)

    release_events = [e for e in events if e.to_release == release]

    for event in release_events:
        # PRESENCE events have a different semantics than the other events - they add a package to a target state
        # only if it had been seen (installed) during the course of the overall target packages
        if event.action == Action.PRESENT:
            for pkg in event.in_pkgs:
                if pkg in seen_pkgs:
                    if pkg in target_pkgs:
                        # Remove the package with the old repository, add the one with the new one
                        target_pkgs.remove(pkg)
                    target_pkgs.add(pkg)
        elif event.action == Action.DEPRECATED:
            if event.in_pkgs.intersection(source_installed_pkgs):
                # Remove packages with old repositories add packages with the new one
                target_pkgs = target_pkgs.difference(event.in_pkgs)
                target_pkgs = target_pkgs.union(event.in_pkgs)
        else:
            # All other packages have the same semantics - they remove their in_pkgs from the system with given
            # from_release and add out_pkgs to the system matching to_release
            are_all_in_pkgs_present = all(in_pkg in source_installed_pkgs for in_pkg in event.in_pkgs)
            is_any_in_pkg_present = any(in_pkg in source_installed_pkgs for in_pkg in event.in_pkgs)

            # For MERGE to be relevant it is sufficient for only one of its in_pkgs to be installed
            if are_all_in_pkgs_present or (event.action == Action.MERGED and is_any_in_pkg_present):
                removed_pkgs = target_pkgs.intersection(event.in_pkgs)
                removed_pkgs_str = ', '.join(str(pkg) for pkg in removed_pkgs) if removed_pkgs else '[]'
                added_pkgs_str = ', '.join(str(pkg) for pkg in event.out_pkgs) if event.out_pkgs else '[]'
                logger.debug('Applying event %d (%s): replacing packages %s with %s',
                             event.id, event.action, removed_pkgs_str, added_pkgs_str)

                # In pkgs are present, event can be applied
                target_pkgs = target_pkgs.difference(event.in_pkgs)
                target_pkgs = target_pkgs.union(event.out_pkgs)

        pkgs_to_demodularize = pkgs_to_demodularize.difference(event.in_pkgs)

    return (target_pkgs, pkgs_to_demodularize)


def remove_undesired_events(events, relevant_to_releases):
    """
    Conservatively remove events that needless, or cause problems for the current implementation:
    - (needless) events with to_release not in relevant releases
    - (problematic) events with the same from_release and the same in_pkgs
    """

    logger = api.current_logger()
    relevant_to_releases = set(relevant_to_releases)

    events_with_same_in_pkgs_and_from_release = defaultdict(list)
    for event in events:
        if event.to_release in relevant_to_releases:
            # NOTE(mhecko): The tuple(sorted(event.in_pkgs))) is ugly, however, the removal of the events with the same
            # #             from_release and in_pkgs is needed only because the current implementation is flawed.
            # #             I would love to rewrite the core algorithm as a "solution to graph reachability problem",
            # #             making the behaviour of PES event scanner purely data driven.
            events_with_same_in_pkgs_and_from_release[(event.from_release, tuple(sorted(event.in_pkgs)))].append(event)

    cleaned_events = []
    for from_release_in_pkgs_pair, problematic_events in events_with_same_in_pkgs_and_from_release.items():
        if len(problematic_events) == 1:
            cleaned_events.append(problematic_events[0])  # There is no problem
            continue

        # E.g., one of the problematic events is to=8.6, other one to=8.7, keep only 8.7
        from_release, dummy_in_pkgs = from_release_in_pkgs_pair
        max_to_release = max((e.to_release for e in problematic_events))
        events_with_max_to_release = [event for event in problematic_events if event.to_release == max_to_release]

        if len(events_with_max_to_release) == 1:
            # If there is a single event with maximal to_release, keep only that
            kept_event = events_with_max_to_release[0]
            event_ids = [event.id for event in problematic_events]
            logger.debug('Events %s have the same in packages and the same from_release %s, keeping %d',
                         event_ids, from_release, kept_event.id)
            cleaned_events.append(kept_event)
            continue

        # There are at least 2 events A, B with the same in_release, out_release and in_pkgs. If A is REMOVE and B
        # performs some conditional mutation (e.g. SPLIT) a race-conflict arises. However, the current
        # implementation would apply these events as `A(input_state) union B(input_state)`, where the input_state
        # is kept immutable. Therefore, B will have an effect regardless of whether A is REMOVAL or not.
        for event in problematic_events:
            cleaned_events.append(event)

    return cleaned_events


def compute_packages_on_target_system(source_pkgs, events, releases):

    seen_pkgs = set(source_pkgs)  # Used to track whether PRESENCE events can be applied
    target_pkgs = set(source_pkgs)

    source_major_version = int(version.get_source_major_version())
    did_processing_cross_major_version = False
    pkgs_to_demodularize = set()  # Modified by compute_pkg_changes

    for release in releases:
        if not did_processing_cross_major_version and release[0] > source_major_version:
            did_processing_cross_major_version = True
            pkgs_to_demodularize = {pkg for pkg in target_pkgs if pkg.modulestream}

        target_pkgs, pkgs_to_demodularize = compute_pkg_changes_between_consequent_releases(target_pkgs, events,
                                                                                            release, seen_pkgs,
                                                                                            pkgs_to_demodularize)
        seen_pkgs = seen_pkgs.union(target_pkgs)

    demodularized_pkgs = {Package(pkg.name, pkg.repository, None) for pkg in pkgs_to_demodularize}
    demodularized_target_pkgs = target_pkgs.difference(pkgs_to_demodularize).union(demodularized_pkgs)

    return (demodularized_target_pkgs, pkgs_to_demodularize)


def compute_rpm_tasks_from_pkg_set_diff(source_pkgs, target_pkgs, pkgs_to_demodularize):
    source_state_pkg_names = {pkg.name for pkg in source_pkgs}
    target_state_pkg_names = {pkg.name for pkg in target_pkgs}

    pkgs_to_install = sorted(target_state_pkg_names.difference(source_state_pkg_names))
    pkgs_to_remove = sorted(source_state_pkg_names.difference(target_state_pkg_names))

    if pkgs_to_install or pkgs_to_remove:
        # NOTE(mhecko): Here we do not want to consider any package that does not have a reference in PES data. There
        # might be missing modularity information, and although the algorithm is correct, trying to enable
        # a non-existent modulestream due to missing modulestream information results in a crash.
        target_pkgs_without_demodularized_pkgs = target_pkgs.difference(pkgs_to_demodularize)

        # Collect the enabled modules as tuples in a set, so we produce every module to reset exactly once
        enabled_modules = {(module.name, module.stream) for module in _get_enabled_modules()}
        modules_to_reset = [Module(name=ms[0], stream=ms[1]) for ms in enabled_modules]

        target_modulestreams = {pkg.modulestream for pkg in target_pkgs_without_demodularized_pkgs if pkg.modulestream}
        modules_to_enable = [Module(name=ms[0], stream=ms[1]) for ms in target_modulestreams]

        return PESRpmTransactionTasks(to_install=pkgs_to_install,
                                      to_remove=pkgs_to_remove,
                                      modules_to_enable=modules_to_enable,
                                      modules_to_reset=modules_to_reset)
    return None


def report_skipped_packages(title, message, skipped_pkgs, remediation=None):
    skipped_pkgs = sorted(skipped_pkgs)

    def make_summary_entry_for_skipped_pkg(pkg):
        entry_template = '- {name}{modulestream} (repoid: {repository})'
        modulestream_str = '' if not pkg.modulestream else '[{}:{}]'.format(*pkg.modulestream)
        return entry_template.format(name=pkg.name, modulestream=modulestream_str, repository=pkg.repository)

    summary = '{} {}\n{}'.format(len(skipped_pkgs),
                                 message,
                                 '\n'.join(make_summary_entry_for_skipped_pkg(pkg) for pkg in skipped_pkgs))
    report_content = [
        reporting.Title(title),
        reporting.Summary(summary),
        reporting.Severity(reporting.Severity.HIGH),
        reporting.Groups([reporting.Groups.REPOSITORY]),
    ]
    if remediation:
        report_content += [reporting.Remediation(hint=remediation)]
    report_content += [reporting.RelatedResource('package', p[0]) for p in skipped_pkgs]
    reporting.create_report(report_content)
    if is_verbose():
        api.current_logger().info(summary)


def remove_new_packages_from_blacklisted_repos(source_pkgs, target_pkgs):
    """
    Remove newly installed packages from blacklisted repositories that were computed to be on the target system.
    """
    blacklisted_repoids = get_blacklisted_repoids()
    new_pkgs = target_pkgs.difference(source_pkgs)
    pkgs_from_blacklisted_repos = set(pkg for pkg in new_pkgs if pkg.repository in blacklisted_repoids)

    if pkgs_from_blacklisted_repos:
        report_skipped_packages(
            title='Packages available in excluded repositories will not be installed',
            message=SKIPPED_PKGS_MSG,
            skipped_pkgs=pkgs_from_blacklisted_repos,
        )
    return blacklisted_repoids, target_pkgs.difference(pkgs_from_blacklisted_repos)


def get_blacklisted_repoids():
    repos_blacklisted = set()
    for blacklist in api.consume(RepositoriesBlacklisted):
        repos_blacklisted.update(blacklist.repoids)
    return repos_blacklisted


def get_enabled_repoids():
    """
    Collects repoids of all enabled repositories on the source system.

    :param repositories_facts: Iterable of RepositoriesFacts containing repositories related info about source system.
    :return: Set of all enabled repository IDs present on the source system.
    :rtype: Set[str]
    """
    enabled_repoids = set()
    for repos in api.consume(RepositoriesFacts):
        for repo_file in repos.repositories:
            for repo in repo_file.data:
                if repo.enabled:
                    enabled_repoids.add(repo.repoid)
    return enabled_repoids


def get_pesid_to_repoid_map(target_pesids):
    """
    Get a dictionary mapping all PESID repositories to their corresponding repoid.

    :param target_pesids: The set of target PES IDs needed to be mapped
    :return: Dictionary mapping the target_pesids to their corresponding repoid
    """

    repositories_map_msgs = api.consume(RepositoriesMapping)
    repositories_map_msg = next(repositories_map_msgs, None)
    if list(repositories_map_msgs):
        api.current_logger().warning('Unexpectedly received more than one RepositoriesMapping message.')
    if not repositories_map_msg:
        raise StopActorExecutionError(
            'Cannot parse RepositoriesMapping data properly',
            details={'Problem': 'Did not receive a message with mapped repositories'}
        )

    rhui_info = next(api.consume(RHUIInfo), None)
    cloud_provider = rhui_info.provider if rhui_info else ''

    repomap = peseventsscanner_repomap.RepoMapDataHandler(repositories_map_msg, cloud_provider=cloud_provider)

    # NOTE: We have to calculate expected target repositories like in the setuptargetrepos actor.
    # It's planned to handle this in different a way in future...

    enabled_repoids = get_enabled_repoids()
    default_channels = peseventsscanner_repomap.get_default_repository_channels(repomap, enabled_repoids)
    repomap.set_default_channels(default_channels)

    exp_pesid_repos = repomap.get_expected_target_pesid_repos(enabled_repoids)
    # FIXME: this is hack now. In case some packages will need a repository
    # with pesid that is not mapped by default regarding the enabled repos,
    # let's use this one representative repository (baseos/appstream) to get
    # data for a guess of the best repository from the requires target pesid..
    # FIXME: this could now fail in case all repos are disabled...
    representative_repo = exp_pesid_repos.get(
        peseventsscanner_repomap.DEFAULT_PESID[version.get_target_major_version()], None
    )
    if not representative_repo:
        api.current_logger().warning('Cannot determine the representative target base repository.')
        api.current_logger().info(
            'Fallback: Create an artificial representative PESIDRepositoryEntry for the repository mapping'
        )
        representative_repo = PESIDRepositoryEntry(
            pesid=peseventsscanner_repomap.DEFAULT_PESID[version.get_target_major_version()],
            arch=api.current_actor().configuration.architecture,
            major_version=version.get_target_major_version(),
            repoid='artificial-repoid',
            repo_type='rpm',
            channel='ga',
            rhui='',
        )

    for pesid in target_pesids:
        if pesid in exp_pesid_repos:
            continue
        # some packages are moved to repos outside of default repomapping
        # try to find the best possible repo for them..
        # FIXME: HACK NOW
        # good way is to modify class to search repo with specific criteria..
        if not representative_repo:
            api.current_logger().warning('Cannot find suitable repository for PES ID: {}'.format(pesid))
            continue
        pesid_repo = repomap._find_repository_target_equivalent(representative_repo, pesid)

        if not pesid_repo:
            api.current_logger().warning('Cannot find suitable repository for PES ID: {}'.format(pesid))
            continue
        exp_pesid_repos[pesid] = pesid_repo

    # Map just pesids to their corresponding repoids ({to_pesid: repoid})
    repositories_mapping = {}
    for pesid, repository in exp_pesid_repos.items():
        if pesid not in target_pesids:
            # We can skip this repo as it was not identified as needed during the processing of PES events
            continue
        if not repository:
            # TODO
            continue
        repositories_mapping[pesid] = repository.repoid

    return repositories_mapping


def replace_pesids_with_repoids_in_packages(packages, source_pkgs_repoids):
    """Replace packages with PESID in their .repository field with ones that have repoid providing the package."""
    # We want to map only PESIDs - if some package had no events, it will its repository set to source system repoid
    packages_with_pesid = {pkg for pkg in packages if pkg.repository not in source_pkgs_repoids}
    packages_without_pesid = packages.difference(packages_with_pesid)

    required_target_pesids = {pkg.repository for pkg in packages_with_pesid}

    pesid_to_repoid_map = get_pesid_to_repoid_map(required_target_pesids)

    packages_without_known_repoid = {pkg for pkg in packages_with_pesid if pkg.repository not in pesid_to_repoid_map}

    if packages_without_known_repoid:
        report_skipped_packages(
            title='Packages from unknown repositories may not be installed',
            message='packages may not be installed or upgraded due to repositories unknown to leapp:',
            skipped_pkgs=packages_without_known_repoid,
            remediation=(
                'In case the listed repositories are mirrors of official repositories for RHEL'
                ' (provided by Red Hat on CDN)'
                ' and their repositories IDs has been customized, you can change'
                ' the configuration to use the official IDs instead of fixing the problem.'
                ' You can also review the projected DNF upgrade transaction result'
                ' in the logs to see what is going to happen, as this does not necessarily mean'
                ' that the listed packages will not be upgraded. You can also'
                ' install any missing packages after the in-place upgrade manually.'
            ),
        )

    packages_with_known_repoid = packages_with_pesid.difference(packages_without_known_repoid)
    packages_with_repoid = {
        Package(p.name, pesid_to_repoid_map[p.repository], p.modulestream) for p in packages_with_known_repoid
    }
    # Packages without pesid are those for which we do not have an event, keep them in target packages
    return packages_with_repoid.union(packages_without_pesid)


def apply_transaction_configuration(source_pkgs):
    source_pkgs_with_conf_applied = set(source_pkgs)
    transaction_configuration = get_transaction_configuration()

    source_pkgs_with_conf_applied = source_pkgs.union(transaction_configuration.to_install)

    # Transaction configuration contains only names of packages to install/remove/keep - there is no modularity
    # information - modify target_pkgs in a way ignoring modulestream information
    pkg_name_to_pkg_info_map = {pkg.name: pkg for pkg in source_pkgs}

    for pkg in transaction_configuration.to_remove:
        if pkg.name in pkg_name_to_pkg_info_map:
            source_pkgs_with_conf_applied.remove(pkg_name_to_pkg_info_map[pkg.name])

    for pkg in transaction_configuration.to_keep:
        if pkg.name in pkg_name_to_pkg_info_map:
            source_pkgs_with_conf_applied.add(pkg_name_to_pkg_info_map[pkg.name])

    return source_pkgs_with_conf_applied


def remove_leapp_related_events(events):
    # NOTE(ivasilev) Need to revisit this once rhel9->rhel10 upgrades become a thing
    leapp_pkgs = rpms.get_leapp_dep_packages(
            major_version=['7', '8']) + rpms.get_leapp_packages(major_version=['7', '8'])
    res = []
    for event in events:
        if not any(pkg.name in leapp_pkgs for pkg in event.in_pkgs):
            res.append(event)
        else:
            api.current_logger().debug('Filtered out leapp related event, event id: {}'.format(event.id))
    return res


def process():
    # Retrieve data - installed_pkgs, transaction configuration, pes events
    events = get_pes_events('/etc/leapp/files', 'pes-events.json')
    if not events:
        return

    releases = get_relevant_releases(events)
    source_pkgs = get_installed_pkgs()
    source_pkgs = apply_transaction_configuration(source_pkgs)

    # Keep track of what repoids have the source packages to be able to determine what are the PESIDs of the computed
    # packages of the target system, so we can distinguish what needs to be repomapped
    repoids_of_source_pkgs = {pkg.repository for pkg in source_pkgs}

    events = remove_leapp_related_events(events)
    events = remove_undesired_events(events, releases)

    # Apply events - compute what packages should the target system have
    target_pkgs, pkgs_to_demodularize = compute_packages_on_target_system(source_pkgs, events, releases)

    # Packages coming out of the events have PESID as their repository, however, we need real repoid
    target_pkgs = replace_pesids_with_repoids_in_packages(target_pkgs, repoids_of_source_pkgs)

    # Apply the desired repository blacklisting
    blacklisted_repoids, target_pkgs = remove_new_packages_from_blacklisted_repos(source_pkgs, target_pkgs)

    # Look at the target packages and determine what repositories to enable
    target_repoids = sorted(set(p.repository for p in target_pkgs) - blacklisted_repoids - repoids_of_source_pkgs)
    repos_to_enable = RepositoriesSetupTasks(to_enable=target_repoids)
    api.produce(repos_to_enable)

    # Compare the packages on source system and the computed packages on target system and determine what to install
    rpm_tasks = compute_rpm_tasks_from_pkg_set_diff(source_pkgs, target_pkgs, pkgs_to_demodularize)
    if rpm_tasks:
        api.produce(rpm_tasks)



import os

import pytest
from six import text_type

from leapp.libraries.actor import checksendmail


@pytest.mark.parametrize('test_input,migrate', [
    ('IPv6:::1\n', True),
    ('IPv6:0:0:0:0:0:0:0:1\n', False),
])
def test_check_migration(tmpdir, monkeypatch, test_input, migrate):
    test_cfg_path = text_type(tmpdir)
    test_cfg_file = os.path.join(test_cfg_path, 'sendmail.cf')
    with open(test_cfg_file, 'w') as file_out:
        file_out.write(test_input)
    monkeypatch.setattr(checksendmail, 'SendmailConfDir', test_cfg_path)
    files = checksendmail.check_files_for_compressed_ipv6()
    if migrate:
        assert files == [test_cfg_file]
    else:
        assert files == []



#!/usr/bin/env python
#
# Tests for bind configuration parsing

from leapp.libraries.common import isccfg

#
# Sample configuration stubs
#
named_conf_default = isccfg.MockConfig("""
//
// named.conf
//
// Provided by Red Hat bind package to configure the ISC BIND named(8) DNS
// server as a caching only nameserver (as a localhost DNS resolver only).
//
// See /usr/share/doc/bind*/sample/ for example named configuration files.
//

options {
    listen-on port 53 { 127.0.0.1; };
    listen-on-v6 port 53 { ::1; };
    directory      "/var/named";
    dump-file      "/var/named/data/cache_dump.db";
    statistics-file "/var/named/data/named_stats.txt";
    memstatistics-file "/var/named/data/named_mem_stats.txt";
    secroots-file   "/var/named/data/named.secroots";
    recursing-file  "/var/named/data/named.recursing";
    allow-query     { localhost; };

    /*
     - If you are building an AUTHORITATIVE DNS server, do NOT enable recursion.
     - If you are building a RECURSIVE (caching) DNS server, you need to enable
       recursion.
     - If your recursive DNS server has a public IP address, you MUST enable access
       control to limit queries to your legitimate users. Failing to do so will
       cause your server to become part of large scale DNS amplification
       attacks. Implementing BCP38 within your network would greatly
       reduce such attack surface
    */
    recursion yes;

    dnssec-enable yes;
    dnssec-validation yes;

    managed-keys-directory "/var/named/dynamic";

    pid-file "/run/named/named.pid";
    session-keyfile "/run/named/session.key";
};

logging {
        channel default_debug {
                file "data/named.run";
                severity dynamic;
        };
};

zone "." IN {
    type hint;
    file "named.ca";
};

# Avoid including files from bind package, may be not installed
# include "/etc/named.rfc1912.zones";
# include "/etc/named.root.key";
include "/dev/null";
""")


options_lookaside_no = isccfg.MockConfig("""
options {
    dnssec-lookaside no;
};
""")


options_lookaside_auto = isccfg.MockConfig("""
options {
    dnssec-lookaside /* no */ auto;
};
""")


options_lookaside_manual = isccfg.MockConfig("""
options {
    # make sure parser handles comments
    dnssec-lookaside "." /* comment to confuse parser */trust-anchor "dlv.isc.org";
};
""")


options_lookaside_commented = isccfg.MockConfig("""
options {
    /* dnssec-lookaside auto; */
};
""")


views_lookaside = isccfg.MockConfig("""
view "v1" IN {
    // This is auto
    dnssec-lookaside auto;
};

options {
    /* This is multi
     * line
     * comment */
    dnssec-lookaside no;
};

view "v2" {
    # Note no IN
    dnssec-lookaside "." trust-anchor "dlv.isc.org";
};
""")

config_empty = isccfg.MockConfig('')

config_empty_include = isccfg.MockConfig('options { include "/dev/null"; };')


def check_in_section(parser, section, key, value):
    """ Helper to check some section was found
        in configuration section and has expected value

        :type parser: IscConfigParser
        :type section: bind.ConfigSection
        :type key: str
        :param value: expected value """
    assert isinstance(section, isccfg.ConfigSection)
    cfgval = parser.find_val_section(section, key)
    assert isinstance(cfgval, isccfg.ConfigSection)
    assert cfgval.value() == value
    return cfgval


def cb_state(statement, state):
    """Callback used in IscConfigParser.walk()"""
    key = statement.var(0).value()
    state[key] = statement


def find_options(parser):
    """Replace IscConfigParser.find_option with walk use"""
    state = {}
    callbacks = {
        'options': cb_state,
    }
    assert len(parser.FILES_TO_CHECK) >= 1
    cfg = parser.FILES_TO_CHECK[0]
    parser.walk(cfg.root_section(), callbacks, state)
    options = state['options']
    if options:
        assert isinstance(options, isccfg.ConfigVariableSection)
        return options.firstblock()
    return None


# End of helpers
#
# Begin of tests


def test_lookaside_no():
    parser = isccfg.IscConfigParser(options_lookaside_no)
    assert len(parser.FILES_TO_CHECK) == 1
    opt = find_options(parser)
    check_in_section(parser, opt, "dnssec-lookaside", "no")


def test_lookaside_commented():
    parser = isccfg.IscConfigParser(options_lookaside_commented)
    assert len(parser.FILES_TO_CHECK) == 1
    opt = find_options(parser)
    assert isinstance(opt, isccfg.ConfigSection)
    lookaside = parser.find_val_section(opt, "dnssec-lookaside")
    assert lookaside is None


def test_default():
    parser = isccfg.IscConfigParser(named_conf_default)
    assert len(parser.FILES_TO_CHECK) >= 2
    opt = find_options(parser)
    check_in_section(parser, opt, "directory", '"/var/named"')
    check_in_section(parser, opt, "session-keyfile", '"/run/named/session.key"')
    check_in_section(parser, opt, "allow-query", '{ localhost; }')
    check_in_section(parser, opt, "recursion", 'yes')
    check_in_section(parser, opt, "dnssec-validation", 'yes')
    check_in_section(parser, opt, "dnssec-enable", 'yes')


def test_key_lookaside():
    parser = isccfg.IscConfigParser(options_lookaside_manual)
    opt = find_options(parser)
    key = parser.find_next_key(opt.config, opt.start+1, opt.end)
    assert isinstance(key, isccfg.ConfigSection)
    assert key.value() == 'dnssec-lookaside'
    value = parser.find_next_val(opt.config, None, key.end+1, opt.end)
    assert value.value() == '"."'
    key2 = parser.find_next_key(opt.config, value.end+1, opt.end)
    assert key2.value() == 'trust-anchor'
    value2a = parser.find_next_val(opt.config, None, key2.end+1, opt.end)
    value2b = parser.find_val(opt.config, 'trust-anchor', value.end+1, opt.end)
    assert value2b.value() == '"dlv.isc.org"'
    assert value2a.value() == value2b.value()
    value3 = parser.find_next_key(opt.config, value2b.end+1, opt.end, end_report=True)
    assert value3.value() == ';'


def test_key_lookaside_all():
    """ Test getting variable arguments after keyword """
    parser = isccfg.IscConfigParser(options_lookaside_manual)
    assert len(parser.FILES_TO_CHECK) == 1
    opt = find_options(parser)
    assert isinstance(opt, isccfg.ConfigSection)
    values = parser.find_values(opt, "dnssec-lookaside")
    assert values is not None
    assert len(values) >= 4
    key = values[0].value()
    assert key == 'dnssec-lookaside'
    assert values[1].value() == '"."'
    assert values[2].value() == 'trust-anchor'
    assert values[3].value() == '"dlv.isc.org"'
    assert values[4].value() == ';'


def test_key_lookaside_simple():
    """ Test getting variable arguments after keyword """
    parser = isccfg.IscConfigParser(options_lookaside_manual)
    assert len(parser.FILES_TO_CHECK) == 1
    stmts = parser.find('options.dnssec-lookaside')
    assert stmts is not None
    assert len(stmts) == 1
    assert isinstance(stmts[0], isccfg.ConfigVariableSection)
    values = stmts[0].values
    assert len(values) >= 4
    key = values[0].value()
    assert key == 'dnssec-lookaside'
    assert values[1].value() == '"."'
    assert values[2].value() == 'trust-anchor'
    assert values[3].value() == '"dlv.isc.org"'
    assert values[4].value() == ';'


def test_find_index():
    """ Test simplified searching for values in sections """
    parser = isccfg.IscConfigParser(named_conf_default)
    assert len(parser.FILES_TO_CHECK) >= 1
    stmts = parser.find('logging.channel.severity')
    assert stmts is not None and len(stmts) == 1
    assert isinstance(stmts[0], isccfg.ConfigVariableSection)
    values = stmts[0].values
    assert len(values) >= 1
    key = values[0].value()
    assert key == 'severity'
    assert values[1].value() == 'dynamic'
    recursion = parser.find('options.recursion')
    assert len(recursion) == 1 and len(recursion[0].values) >= 2
    assert recursion[0].values[0].value() == 'recursion'
    assert recursion[0].values[1].value() == 'yes'


def cb_view(statement, state):
    if 'view' not in state:
        state['view'] = {}
    name = statement.var(1).invalue()
    second = statement.var(2)
    if second.type() != isccfg.ConfigSection.TYPE_BLOCK:
        name = second.value() + '_' + name
    state['view'][name] = statement


def test_key_views_lookaside():
    """ Test getting variable arguments for views """

    parser = isccfg.IscConfigParser(views_lookaside)
    assert len(parser.FILES_TO_CHECK) == 1
    opt = find_options(parser)
    assert isinstance(opt, isccfg.ConfigSection)
    opt_val = parser.find_values(opt, "dnssec-lookaside")
    assert isinstance(opt_val[1], isccfg.ConfigSection)
    assert opt_val[1].value() == 'no'

    state = {}
    callbacks = {
        'view': cb_view,
    }
    assert len(parser.FILES_TO_CHECK) >= 1
    cfg = parser.FILES_TO_CHECK[0]
    parser.walk(cfg.root_section(), callbacks, state)

    views = state['view']
    assert len(views) == 2

    v1 = views['IN_v1']
    assert isinstance(v1, isccfg.ConfigVariableSection)
    v1b = v1.firstblock()
    assert isinstance(v1b, isccfg.ConfigSection)
    v1_la = parser.find_val_section(v1b, "dnssec-lookaside")
    assert isinstance(v1_la, isccfg.ConfigSection)
    assert v1_la.value() == 'auto'

    v2 = views['v2']
    assert isinstance(v2, isccfg.ConfigVariableSection)
    v2b = v2.firstblock()
    assert isinstance(v2b, isccfg.ConfigSection)
    v2_la = parser.find_values(v2b, "dnssec-lookaside")
    assert isinstance(v2_la[1], isccfg.ConfigSection)
    assert v2_la[1].value() == '"."'
    assert isinstance(v2_la[3], isccfg.ConfigSection)
    assert v2_la[3].value() == '"dlv.isc.org"'


def test_remove_comments():
    """ Test removing comments works as expected """

    parser = isccfg.IscConfigParser(views_lookaside)
    assert len(parser.FILES_TO_CHECK) == 1
    cfg = parser.FILES_TO_CHECK[0]
    assert isinstance(cfg, isccfg.ConfigFile)
    removed_comments = parser._remove_comments(cfg.buffer)
    assert len(removed_comments) < len(cfg.buffer)
    replaced_comments = parser._replace_comments(cfg.buffer)
    assert len(replaced_comments) == len(cfg.buffer)
    assert 'This is auto' not in replaced_comments
    assert 'comment' not in replaced_comments
    assert 'Note no IN' not in replaced_comments


def test_walk():
    """ Test walk function of parser """

    callbacks = {
        'options': cb_state,
        'dnssec-lookaside': cb_state,
        'dnssec-validation': cb_state,
    }
    state = {}
    parser = isccfg.IscConfigParser(views_lookaside)
    assert len(parser.FILES_TO_CHECK) == 1
    cfg = parser.FILES_TO_CHECK[0]
    parser.walk(cfg.root_section(), callbacks, state)
    assert 'options' in state
    assert 'dnssec-lookaside' in state
    assert 'dnssec-validation' not in state


def test_empty_config():
    """ Test empty configuration """

    callbacks = {}

    parser = isccfg.IscConfigParser(config_empty)
    assert len(parser.FILES_TO_CHECK) == 1
    cfg = parser.FILES_TO_CHECK[0]
    parser.walk(cfg.root_section(), callbacks)
    assert cfg.buffer == ''


def test_empty_include_config():
    """ Test empty configuration """

    callbacks = {}

    parser = isccfg.IscConfigParser(config_empty_include)
    assert len(parser.FILES_TO_CHECK) == 2
    cfg = parser.FILES_TO_CHECK[0]
    parser.walk(cfg.root_section(), callbacks)
    assert cfg.buffer == 'options { include "/dev/null"; };'

    null_cfg = parser.FILES_TO_CHECK[1]
    parser.walk(null_cfg.root_section(), callbacks)
    assert null_cfg.buffer == ''


if __name__ == '__main__':
    test_key_views_lookaside()



from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import InstalledRPM


def get_installed_desktops():
    api.current_logger().info("  Detecting desktop environments  ")
    api.current_logger().info("==================================")

    # Detect installed desktops by one of the base rpm packages
    kde_desktop_installed = has_package(InstalledRPM, "kde-workspace")
    gnome_desktop_installed = has_package(InstalledRPM, "gnome-session")
    api.current_logger().info("* KDE installed: {0}".format(kde_desktop_installed))
    api.current_logger().info("* Gnome installed: {0}".format(gnome_desktop_installed))
    api.current_logger().info("----------------------------------")

    return {"gnome_installed": gnome_desktop_installed, "kde_installed": kde_desktop_installed}



import textwrap

from leapp.libraries.actor.removeoldpammodulesscanner import RemoveOldPAMModulesScannerLibrary
from leapp.libraries.common.pam import PAM


def get_config(config):
    return textwrap.dedent(config).strip()


def test_RemoveOldPAMModulesScannerLibrary_process__pkcs11():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_pkcs11.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = RemoveOldPAMModulesScannerLibrary(PAM(pam))
    model = obj.process()
    assert model.modules == ['pam_pkcs11']


def test_RemoveOldPAMModulesScannerLibrary_process__krb5():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_krb5.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = RemoveOldPAMModulesScannerLibrary(PAM(pam))
    model = obj.process()
    assert model.modules == ['pam_krb5']


def test_RemoveOldPAMModulesScannerLibrary_process__all():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_krb5.so
    auth sufficient pam_pkcs11.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = RemoveOldPAMModulesScannerLibrary(PAM(pam))
    model = obj.process()
    assert len(model.modules) == 2
    assert 'pam_krb5' in model.modules
    assert 'pam_pkcs11' in model.modules


def test_RemoveOldPAMModulesScannerLibrary_process__none():
    pam = get_config('''
    auth sufficient pam_unix.so
    auth sufficient pam_sss.so
    auth required pam_deny.so
    ''')

    obj = RemoveOldPAMModulesScannerLibrary(PAM(pam))
    model = obj.process()
    assert not model.modules



import pytest

from leapp.libraries.actor.cupsscanner import directive_exists

testdata = (
    ('PrintcapFormat', 'ble', False),
    ('PrintcapFormat', '', False),
    ('PrintcapFormat', '#PrintcapFormat', False),
    ('PrintcapFormat', 'PrintcapFormat', True),
    ('PrintcapFormat', '    PrintcapFormat', True)
)


@pytest.mark.parametrize("string, line, expected", testdata)
def test_directive_exists(string, line, expected):

    ret = directive_exists(string, line)

    assert ret == expected



import pytest

from leapp import reporting
from leapp.libraries.actor import checkopensslconf
from leapp.libraries.common.config import architecture
from leapp.libraries.common.testutils import create_report_mocked, CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import api
from leapp.models import DistributionSignedRPM, FileInfo, RPM, TrackedFilesInfoSource

_DUMP_PKG_NAMES = ['random', 'pkgs', 'openssl-ibmca-nope', 'ibmca', 'nope-openssl-ibmca']
_SSL_CONF = checkopensslconf.DEFAULT_OPENSSL_CONF


def _msg_pkgs(pkgnames):
    rpms = []
    for pname in pkgnames:
        rpms.append(RPM(
            name=pname,
            epoch='0',
            version='1.0',
            release='1',
            arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51',
            packager='Red Hat, Inc. (auxiliary key 2) <security@redhat.com>'

        ))
    return DistributionSignedRPM(items=rpms)


@pytest.mark.parametrize('arch,pkgnames,ibmca_report', (
    (architecture.ARCH_S390X, [], False),
    (architecture.ARCH_S390X, _DUMP_PKG_NAMES, False),
    (architecture.ARCH_S390X, ['openssl-ibmca'], True),
    (architecture.ARCH_S390X, _DUMP_PKG_NAMES + ['openssl-ibmca'], True),
    (architecture.ARCH_S390X, ['openssl-ibmca'] + _DUMP_PKG_NAMES, True),

    # stay false for non-IBM-z arch - invalid scenario basically
    (architecture.ARCH_X86_64, ['openssl-ibmca'], False),
    (architecture.ARCH_PPC64LE, ['openssl-ibmca'], False),
    (architecture.ARCH_ARM64, ['openssl-ibmca'], False),

))
@pytest.mark.parametrize('src_maj_ver', ('7', '8', '9'))
def test_check_ibmca(monkeypatch, src_maj_ver, arch, pkgnames, ibmca_report):
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        arch=arch,
        msgs=[_msg_pkgs(pkgnames)],
        src_ver='{}.6'.format(src_maj_ver),
        dst_ver='{}.0'.format(int(src_maj_ver) + 1)
    ))
    checkopensslconf.check_ibmca()

    if not ibmca_report:
        assert not reporting.create_report.called, 'IBMCA report created when it should not.'
    else:
        assert reporting.create_report.called, 'IBMCA report has not been created.'


def _msg_files(fnames_changed, fnames_untouched):
    res = []
    for fname in fnames_changed:
        res.append(FileInfo(
            path=fname,
            exists=True,
            is_modified=True
        ))

    for fname in fnames_untouched:
        res.append(FileInfo(
            path=fname,
            exists=True,
            is_modified=False
        ))

    return TrackedFilesInfoSource(files=res)


# NOTE(pstodulk): Ignoring situation when _SSL_CONF is missing (modified, do not exists).
# It's not a valid scenario actually, as this file just must exists on the system to
# consider it in a supported state.
@pytest.mark.parametrize('msg,openssl_report', (
    # matrix focused on openssl reports only (positive)
    (_msg_files([], []), False),
    (_msg_files([_SSL_CONF], []), True),
    (_msg_files(['what/ever', _SSL_CONF, 'something'], []), True),
    (_msg_files(['what/ever'], [_SSL_CONF]), False),
))
@pytest.mark.parametrize('src_maj_ver', ('7', '8', '9'))
def test_check_openssl(monkeypatch, src_maj_ver, msg, openssl_report):
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked(
        msgs=[msg],
        src_ver='{}.6'.format(src_maj_ver),
        dst_ver='{}.0'.format(int(src_maj_ver) + 1)
    ))
    checkopensslconf.process()

    if not openssl_report:
        assert not reporting.create_report.called, 'OpenSSL report created when it should not.'
    else:
        assert reporting.create_report.called, 'OpenSSL report has not been created.'



import os

import pytest

from leapp.libraries.actor import scanner
from leapp.libraries.common.config import architecture, version
from leapp.models import GrubConfigError, Report

CUR_DIR = os.path.dirname(os.path.abspath(__file__))


def test_correct_config_error_detection():
    assert not scanner.detect_config_error(os.path.join(CUR_DIR, 'files/error_detection/grub.correct'))
    assert not scanner.detect_config_error(os.path.join(CUR_DIR, 'files/error_detection/grub.correct_trailing_space'))
    assert not scanner.detect_config_error(os.path.join(CUR_DIR, 'files/error_detection/grub.correct_comment'))
    assert not scanner.detect_config_error(os.path.join(CUR_DIR, 'files/error_detection/grub.correct_puppet'))


def test_wrong_config_error_detection():
    assert scanner.detect_config_error(os.path.join(CUR_DIR, 'files/error_detection/grub.wrong'))
    assert scanner.detect_config_error(os.path.join(CUR_DIR, 'files/error_detection/grub.wrong1'))


def test_all_errors_produced(current_actor_context, monkeypatch):
    # Tell the actor we are not running on s390x
    monkeypatch.setattr(architecture, 'matches_architecture', lambda _: False)
    monkeypatch.setattr(version, 'get_source_version', lambda: '7.9')
    # Set that all checks failed
    monkeypatch.setattr(scanner, 'is_grub_config_missing_final_newline', lambda _: True)
    monkeypatch.setattr(scanner, 'is_grubenv_corrupted', lambda _: True)
    monkeypatch.setattr(scanner, 'detect_config_error', lambda _: True)
    # Run the actor
    current_actor_context.run()
    # Check that exactly 3 messages of different types are produced
    errors = current_actor_context.consume(GrubConfigError)
    assert len(errors) == 3
    for err_type in [GrubConfigError.ERROR_MISSING_NEWLINE, GrubConfigError.ERROR_CORRUPTED_GRUBENV,
                     GrubConfigError.ERROR_GRUB_CMDLINE_LINUX_SYNTAX]:
        distinct_error = next((e for e in errors if e.error_type == err_type), None)
        assert distinct_error
        assert distinct_error.files


@pytest.mark.parametrize(
    ('config_contents', 'error_detected'),
    [
        ('GRUB_DEFAULT=saved\nGRUB_DISABLE_SUBMENU=true\n', False),
        ('GRUB_DEFAULT=saved\nGRUB_DISABLE_SUBMENU=true', True)
    ]
)
def test_is_grub_config_missing_final_newline(monkeypatch, config_contents, error_detected):

    config_path = '/etc/default/grub'

    def mocked_get_config_contents(path):
        assert path == config_path
        return config_contents

    monkeypatch.setattr(scanner, '_get_config_contents', mocked_get_config_contents)
    assert scanner.is_grub_config_missing_final_newline(config_path) == error_detected


def test_correct_config_corrupted_grubenv():
    assert not scanner.is_grubenv_corrupted(os.path.join(CUR_DIR, 'files/corrupted_grubenv/grubenv.correct'))


def test_wrong_config_corrupted_grubenv():
    assert scanner.is_grubenv_corrupted(os.path.join(CUR_DIR, 'files/corrupted_grubenv/grubenv.wrong1'))
    assert scanner.is_grubenv_corrupted(os.path.join(CUR_DIR, 'files/corrupted_grubenv/grubenv.wrong2'))



import os

from leapp import reporting
from leapp.libraries.actor import checkosrelease
from leapp.libraries.common.config import version
from leapp.libraries.common.testutils import create_report_mocked, produce_mocked
from leapp.utils.report import is_inhibitor


def test_skip_check(monkeypatch):
    monkeypatch.setenv('LEAPP_DEVEL_SKIP_CHECK_OS_RELEASE', '1')
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    assert checkosrelease.skip_check()
    assert reporting.create_report.called == 1
    assert 'Skipped OS release check' in reporting.create_report.report_fields['title']
    assert reporting.create_report.report_fields['severity'] == 'high'


def test_no_skip_check(monkeypatch):
    monkeypatch.delenv('LEAPP_DEVEL_SKIP_CHECK_OS_RELEASE', raising=False)
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    assert not checkosrelease.skip_check()
    assert reporting.create_report.called == 0


def test_not_supported_release(monkeypatch):
    monkeypatch.setattr(version, "is_supported_version", lambda: False)
    monkeypatch.setattr(version, "get_source_major_version", lambda: '7')
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    checkosrelease.check_os_version()
    assert reporting.create_report.called == 1
    assert 'The installed OS version is not supported' in reporting.create_report.report_fields['title']
    assert is_inhibitor(reporting.create_report.report_fields)


def test_supported_release(monkeypatch):
    monkeypatch.setattr(version, "is_supported_version", lambda: True)
    monkeypatch.setattr(reporting, "create_report", create_report_mocked())

    checkosrelease.check_os_version()
    assert reporting.create_report.called == 0



from leapp.models import fields, Model
from leapp.topics import SystemInfoTopic


class CurrentKernel(Model):
    topic = SystemInfoTopic
    version = fields.String()
    release = fields.String()
    arch = fields.String()



import os

from leapp.libraries.stdlib import api, CalledProcessError, run


def is_mdraid_dev(dev):
    """
    Check if a given device is an md (Multiple Device) device

    It is expected that the "mdadm" command is available,
    if it's not it is assumed the device is not an md device.

    :return: True if the device is an md device, False otherwise
    :raises CalledProcessError: If an error occurred
    """
    fail_msg = 'Could not check if device "{}" is an md device: {}'
    if not os.path.exists('/usr/sbin/mdadm'):
        api.current_logger().warning(fail_msg.format(
            dev, '/usr/sbin/mdadm is not installed.'
        ))
        return False
    try:
        result = run(['mdadm', '--query', dev])
    except CalledProcessError as err:
        err.message = fail_msg.format(dev, err)
        raise  # let the calling actor handle the exception

    return '--detail' in result['stdout']


def get_component_devices(raid_dev):
    """
    Get list of component devices in an md (Multiple Device) array

    :return: The list of component devices or None in case of error
    :raises ValueError: If the device is not an mdraid device
    """
    try:
        # using both --verbose and --brief for medium verbosity
        result = run(['mdadm', '--detail', '--verbose', '--brief', raid_dev])
    except (OSError, CalledProcessError) as err:
        api.current_logger().warning(
            'Could not get md array component devices: {}'.format(err)
        )
        return None
    # example output:
    # ARRAY /dev/md0 level=raid1 num-devices=2 metadata=1.2 name=localhost.localdomain:0 UUID=c4acea6e:d56e1598:91822e3f:fb26832c # noqa: E501; pylint: disable=line-too-long
    #     devices=/dev/vda1,/dev/vdb1
    if 'does not appear to be an md device' in result['stdout']:
        raise ValueError("Expected md device, but got: {}".format(raid_dev))

    return sorted(result['stdout'].rsplit('=', 2)[-1].strip().split(','))



from leapp.libraries.actor import selinuxcontentscanner
from leapp.libraries.common.config import version
from leapp.libraries.stdlib import CalledProcessError


class run_mocked(object):
    def __init__(self):
        self.args = []
        self.called = 0

    def __call__(self, args, split=True):
        self.called += 1
        self.args = args

        if self.args == ['semodule', '-lfull']:
            stdout = ["400 permissive_abrt_t cil",
                      "400   zebra       cil",
                      "300 zebra       cil",
                      "100 vpn               pp  ",
                      "099 zebra             cil     ",
                      "100   minissdpd         pp"]

        elif self.args == ['semanage', 'export']:
            stdout = ["boolean -D",
                      "login -D",
                      "interface -D",
                      "user -D",
                      "port -D",
                      "node -D",
                      "fcontext -D",
                      "module -D",
                      "boolean -m -1 cron_can_relabel",
                      "port -a -t http_port_t -p udp 81",
                      "fcontext -a -f a -t httpd_sys_content_t '/web(/.*)?'",
                      "fcontext -a -f a -t cgdcbxd_exec_t '/ganesha(/.*)?'"]

        return {'stdout': stdout}


class run_mocked_fail(object):
    def __init__(self):
        self.called = 0

    def __call__(self, args, split=True):
        raise CalledProcessError(self, 1, "Mock error ;)")


def test_list_selinux_modules(monkeypatch):
    monkeypatch.setattr(selinuxcontentscanner, "run", run_mocked())

    assert selinuxcontentscanner.list_selinux_modules() == [
        ("permissive_abrt_t", "400"),
        ("zebra", "400"),
        ("zebra", "300"),
        ("vpn", "100"),
        ("zebra", "099"),
        ("minissdpd", "100"),
    ]

    monkeypatch.setattr(selinuxcontentscanner, "run", run_mocked_fail())

    assert selinuxcontentscanner.list_selinux_modules() == []


def test_get_selinux_customizations(monkeypatch):
    monkeypatch.setattr(version, "get_source_major_version", lambda: '8')
    monkeypatch.setattr(selinuxcontentscanner, "run", run_mocked())

    (semanage_valid, semanage_removed) = selinuxcontentscanner.get_selinux_customizations()

    assert len(semanage_valid) == 3
    assert semanage_valid[0] == "boolean -m -1 cron_can_relabel"
    assert semanage_valid[1] == "port -a -t http_port_t -p udp 81"
    assert semanage_valid[2] == "fcontext -a -f a -t httpd_sys_content_t '/web(/.*)?'"
    assert semanage_removed == ["fcontext -a -f a -t cgdcbxd_exec_t '/ganesha(/.*)?'"]



from leapp.actors import Actor
from leapp.libraries.actor.redhatsignedrpmcheck import check_unsigned_packages
from leapp.models import InstalledUnsignedRPM
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class RedHatSignedRpmCheck(Actor):
    """
    Check if there are packages not signed by Red Hat in use. If yes, warn user about it.

    If any any installed RPM package does not contain a valid signature from Red Hat, a message
    containing a warning is produced.
    """

    name = 'red_hat_signed_rpm_check'
    consumes = (InstalledUnsignedRPM,)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        check_unsigned_packages()



from leapp.actors import Actor
from leapp.libraries.actor import scankernel
from leapp.models import InstalledTargetKernelInfo, InstalledTargetKernelVersion, KernelInfo, TransactionCompleted
from leapp.tags import IPUWorkflowTag, RPMUpgradePhaseTag


class ScanInstalledTargetKernelVersion(Actor):
    """
    Scan for the version of the newly installed kernel

    This actor will query rpm for all kernel-core packages and reports the
    first matching target system kernel RPM. In case the RHEL Real Time has
    been detected on the original system, the kernel-rt-core rpm is searched.
    If the rpm is missing, fallback for standard kernel RPM.
    """

    name = 'scan_installed_target_kernel_version'
    consumes = (TransactionCompleted, KernelInfo)
    produces = (InstalledTargetKernelInfo, InstalledTargetKernelVersion)
    tags = (RPMUpgradePhaseTag, IPUWorkflowTag)

    def process(self):
        scankernel.process()



from leapp.models import fields, Model
from leapp.topics import BootPrepTopic, SystemInfoTopic
from leapp.utils.deprecation import deprecated


class DracutModule(Model):
    """
    Specify a dracut module that should be included into the initramfs

    The specified dracut module has to be compatible with the target system.

    See the description of UpgradeInitramfsTasks and TargetInitramfsTasks
    for more information about the role of initramfs in the in-place upgrade
    process.
    """
    topic = BootPrepTopic

    name = fields.String()
    """
    Name of the dracut module that should be added (--add option of dracut)
    when a initramfs is built.
    """

    module_path = fields.Nullable(fields.String(default=None))
    """
    module_path specifies dracut modules that are supposed to be copied

    If the path is not set, the given name will just be activated. IOW,
    if the dracut module is stored outside the /usr/lib/dracut/modules.d/
    directory, set the absolute path to it, so leapp will manage it during
    the upgrade to ensure the module will be added into the initramfs.

    The module has to be stored on the local storage. In such a case, it is
    recommended to store it into the 'files' directory of an actor generating
    this object.

    Note: It's expected to set the full path from the host POV. In case
    of actions inside containers, the module is still copied from the HOST
    into the container workspace.
    """


class KernelModule(Model):
    """
    Specify a kernel module that should be included into the initramfs

    The specified kernel module has to be compatible with the target system.

    See the description of UpgradeInitramfsTasks and TargetInitramfsTasks
    for more information about the role of initramfs in the in-place upgrade
    process.
    """
    topic = BootPrepTopic

    name = fields.String()
    """
    The kernel module that should be added (--add-drivers option of dracut)
    when a initramfs is built. The possible options are

    1. ``=<kernel subdir>[/<kernel subdir>...]`` like ``=drivers/hid``
    2. ``<module name>``
    """

    module_path = fields.Nullable(fields.String(default=None))
    """
    module_path specifies kernel modules that are supposed to be copied

    If the path is not set, the given name will just be activated. IOW,
    if the kernel module is stored outside the /usr/lib/modules/$(uname -r)/
    directory, set the absolute path to it, so leapp will manage it during
    the upgrade to ensure the module will be added into the initramfs.

    The module has to be stored on the local storage mounted in a persistent
    fashion (/etc/fstab entry). In such a case, it is recommended to store it
    into the 'files' directory of an actor generating this object.

    Note: It's expected to set the full path from the host POV. In case
    of actions inside containers, the module is still copied from the HOST
    into the container workspace.
    """


class UpgradeInitramfsTasks(Model):
    """
    Influence generating of the (leapp) upgrade initramfs

    The upgrade initramfs is used during the crucial part of the upgrade,
    in which the original rpms are upgraded, configuration of applications
    are migrated, etc. To be able to boot into the leapp upgrade environment
    correctly, it is expected all needed drivers, configuration files, ... are
    included inside the upgrade initramfs. Produce this message with
    expected content to influence the upgrade initramfs.

    If some specific rpms or content is required to be able to build the
    upgrade initramfs, see the <container-model>.

    Note: The built initramfs is composed of stuff for the target system.
    In example, if you are on RHEL 7 and plan the upgrade to RHEL 8, you need
    to provide content (e.g. drivers, dracut modules) compatible with
    RHEL 8 system.
    """
    topic = BootPrepTopic

    include_files = fields.List(fields.String(), default=[])
    """
    List of files (canonical filesystem paths) to include in the initramfs
    """

    include_dracut_modules = fields.List(fields.Model(DracutModule), default=[])
    """
    List of dracut modules that should be installed in the initramfs.

    See the DracutModule model for more information.
    """

    include_kernel_modules = fields.List(fields.Model(KernelModule), default=[])
    """
    List of kernel modules that should be installed in the initramfs.

    See the KernelModule model for more information.
    """


class TargetInitramfsTasks(UpgradeInitramfsTasks):
    """
    Analogy to UpgradeInitramfsTasks, but referring to the target initram disk.

    Target initramfs is the one, that will be used to boot to your upgraded
    system. If you want to ensure that you are able to boot into the target
    (upgraded) system, it is possible you need to add same stuff as you added
    into the upgrade initramfs.

    If some specific rpms are required to be able to build the upgrade
    initramfs, install these via the RpmTransactionTasks model.
    """


@deprecated(since='2021-10-10', message='Replaced by TargetInitramfsTasks.')
class InitrdIncludes(Model):
    """
    List of files (canonical filesystem paths) to include in RHEL-8 initramfs
    """
    topic = SystemInfoTopic

    files = fields.List(fields.String())


@deprecated(since='2021-10-10', message='Replaced by UpgradeInitramfsTasks.')
class UpgradeDracutModule(Model):
    """
    Specify a dracut module that should be included into the (leapp) upgrade initramfs.

    The upgrade initramfs is used during the crucial part of the upgrade,
    in which the original rpms are upgraded. If a dracut module is required to
    be included inside the upgrade initramfs (e.g. because it is needed
    to handle/initialize your storage properly), produce this msg.
    """
    topic = BootPrepTopic

    name = fields.String()
    """
    Name of the dracut module that should be added (--add option of dracut)
    """

    module_path = fields.Nullable(fields.String(default=None))
    """
    module_path specifies dracut modules that are to be copied

    If the path is not set, the given name will just be activated.
    """



import os
import shutil

from leapp.libraries.stdlib import api, CalledProcessError, run

# dict(orig_dir: new_dir)
DIRS_CHANGE = {
    '/etc/pki/ca-trust/source/blacklist/': '/etc/pki/ca-trust/source/blocklist/',
    '/usr/share/pki/ca-trust-source/blacklist/': '/usr/share/pki/ca-trust-source/blocklist/'
}


def _link_src_path(filepath):
    """
    Return expected target path for the symlink.

    In case the symlink points to one of dirs supposed to be migrated in this
    actor, we need to point to the new directory instead.

    In case the link points anywhere else, keep the target path as it is.
    """
    realpath = os.path.realpath(filepath)
    for dirname in DIRS_CHANGE:
        if realpath.startswith(dirname):
            return realpath.replace(dirname, DIRS_CHANGE[dirname])

    # it seems we can keep this path
    return realpath


def _migrate_file(filename, src_basedir):
    dst_path = filename.replace(src_basedir, DIRS_CHANGE[src_basedir])
    if os.path.exists(dst_path):
        api.current_logger().info(
            'Skipping migration of the {} certificate. The target file already exists'
            .format(filename)
        )
        return
    os.makedirs(os.path.dirname(dst_path), mode=0o755, exist_ok=True)
    if os.path.islink(filename):
        # create the new symlink instead of the moving the file
        # as the target path could be different as well
        link_src_path = _link_src_path(filename)
        # TODO: is the broken symlink ok?
        os.symlink(link_src_path, dst_path)
        os.unlink(filename)
    else:
        # normal file, just move it
        shutil.move(filename, dst_path)


def _get_files(dirname):
    return run(['find', dirname, '-type', 'f,l'], split=True)['stdout']


def process():
    for dirname in DIRS_CHANGE:
        if not os.path.exists(dirname):
            # The directory does not exist; nothing to do here
            continue
        try:
            blacklisted_certs = _get_files(dirname)
        except (CalledProcessError, OSError) as e:
            # TODO: create post-upgrade report
            api.current_logger().error('Cannot get list of files in {}: {}.'.format(dirname, e))
            api.current_logger().error('Certificates under {} must be migrated manually.'.format(dirname))
            continue
        failed_files = []
        for filename in blacklisted_certs:
            try:
                _migrate_file(filename, dirname)
            except OSError as e:
                api.current_logger().error(
                    'Failed migration of blacklisted certificate {}: {}'
                    .format(filename, e)
                )
                failed_files.append(filename)
        if not failed_files:
            # the failed removal is not such a big issue here
            # clean the dir if all files have been migrated successfully
            shutil.rmtree(dirname, ignore_errors=True)
    try:
        run(['/usr/bin/update-ca-trust'])
    except (CalledProcessError, OSError) as e:
        api.current_logger().error(
            'Cannot update CA trust on the system.'
            ' It needs to be done manually after the in-place upgrade.'
            ' Reason: {}'.format(e)
        )



from leapp.models import DistributionSignedRPM, RPM
from leapp.reporting import Report
from leapp.snactor.fixture import current_actor_context

RH_PACKAGER = 'Red Hat, Inc. <http://bugzilla.redhat.com/bugzilla>'


def test_actor_with_grep_package(current_actor_context):
    rpms = [
        RPM(name='wireshark', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
        RPM(name='powertop', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')]

    current_actor_context.feed(DistributionSignedRPM(items=rpms))
    current_actor_context.run()
    assert current_actor_context.consume(Report)


def test_actor_without_grep_package(current_actor_context):
    rpms = [
        RPM(name='powertop', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51'),
        RPM(name='sed', version='0.1', release='1.sm01', epoch='1', packager=RH_PACKAGER, arch='noarch',
            pgpsig='RSA/SHA256, Mon 01 Jan 1970 00:00:00 AM -03, Key ID 199e2f91fd431d51')]

    current_actor_context.feed(DistributionSignedRPM(items=rpms))
    current_actor_context.run()
    assert not current_actor_context.consume(Report)



from leapp.actors import Actor
from leapp.libraries.actor.mariadbcheck import report_installed_packages
from leapp.models import DistributionSignedRPM, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class MariadbCheck(Actor):
    """
    Actor checking for presence of MariaDB installation.

    Provides user with information related to upgrading systems
    with MariaDB installed.
    """
    name = 'mariadb_check'
    consumes = (DistributionSignedRPM,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        report_installed_packages()



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class XorgDrv(Model):
    """
    Name of the Xorg driver in use and whether it has custom options set.

    This model is not expected to be used as a message (produced/consumed by actors).
    It is used from within the XorgDrvFacts model.
    """
    topic = SystemFactsTopic

    driver = fields.String()
    has_options = fields.Boolean(default=False)


class XorgDrvFacts(Model):
    """
    List of Xorg drivers.
    """
    topic = SystemFactsTopic

    xorg_drivers = fields.List(fields.Model(XorgDrv))



import contextlib
import functools
import os
import re
import time

from leapp import reporting
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.common import repofileutils
from leapp.libraries.common.config import get_env
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import RHSMInfo

_RE_REPO_UID = re.compile(r'Repo ID:\s*([^\s]+)')
_RE_RELEASE = re.compile(r'Release:\s*([^\s]+)')
_RE_SKU_CONSUMED = re.compile(r'SKU:\s*([^\s]+)')
_ATTEMPTS = 5
_RETRY_SLEEP = 5
_DEFAULT_RHSM_REPOFILE = '/etc/yum.repos.d/redhat.repo'

SCA_TEXT = "Content Access Mode is set to Simple Content Access"


def _rhsm_retry(max_attempts, sleep=None):
    """
    A decorator to retry executing a function/method if unsuccessful.

    The function/method execution is considered unsuccessful when it raises StopActorExecutionError.

    :param max_attempts: Maximum number of attempts to execute the decorated function/method.
    :param sleep: Time to wait between attempts. In seconds.
    """
    def impl(f):
        @functools.wraps(f)
        def wrapper(*args, **kwargs):
            attempts = 0
            while True:
                attempts += 1
                try:
                    return f(*args, **kwargs)
                except StopActorExecutionError:
                    if max_attempts <= attempts:
                        api.current_logger().warning(
                            'Attempt %d of %d to perform %s failed. Maximum number of retries has been reached.',
                            attempts, max_attempts, f.__name__)
                        raise
                    if sleep:
                        api.current_logger().info(
                            'Attempt %d of %d to perform %s failed - Retrying after %s seconds',
                            attempts, max_attempts, f.__name__, str(sleep))
                        time.sleep(sleep)
                    else:
                        api.current_logger().info(
                            'Attempt %d of %d to perform %s failed - Retrying...', attempts, max_attempts, f.__name__)
        return wrapper
    return impl


@contextlib.contextmanager
def _handle_rhsm_exceptions(hint=None):
    """
    Context manager based function that handles exceptions of `run` for the subscription-manager calls.
    """
    try:
        yield
    except OSError as e:
        api.current_logger().error('Failed to execute subscription-manager executable')
        raise StopActorExecutionError(
            message='Unable to execute subscription-manager executable: {}'.format(str(e)),
            details={
                'hint': 'Please ensure subscription-manager is installed and executable.'
            }
        )
    except CalledProcessError as e:
        _def_hint = (
            'Please ensure you have a valid RHEL subscription and your network is up.'
            ' If you are using proxy for Red Hat subscription-manager, please make sure'
            ' it is specified inside the /etc/rhsm/rhsm.conf file.'
            ' Or use the --no-rhsm option when running leapp, if you do not want to'
            ' use subscription-manager for the in-place upgrade and you want to'
            ' deliver all target repositories by yourself or using RHUI on public cloud.'
        )
        raise StopActorExecutionError(
            message='A subscription-manager command failed to execute',
            details={
                'details': str(e),
                'stderr': e.stderr,
                'hint': hint or _def_hint
            }
        )


def skip_rhsm():
    """Check whether we should skip RHSM related code."""
    return get_env('LEAPP_NO_RHSM', '0') == '1'


def with_rhsm(f):
    """Decorator to allow skipping RHSM functions by executing a no-op."""
    @functools.wraps(f)
    def wrapper(*args, **kwargs):
        if not skip_rhsm():
            return f(*args, **kwargs)
        return None
    return wrapper


@with_rhsm
def get_attached_skus(context):
    """
    Retrieve the list of the SKUs the system is attached to with the subscription-manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: SKUs the current system is attached to.
    :rtype: List(string)
    """
    with _handle_rhsm_exceptions():
        result = context.call(['subscription-manager', 'list', '--consumed'], split=False)
        return _RE_SKU_CONSUMED.findall(result['stdout'])


@with_rhsm
def get_rhsm_status(context):
    """
    Retrieve "subscription-manager status" output.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: "subscription manager status" output
    :rtype: String
    """
    with _handle_rhsm_exceptions():
        result = context.call(['subscription-manager', 'status'])
        return result['stdout']


def is_manifest_sca(context):
    """
    Check if SCA manifest is used in Satellite.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: True if SCA manifest is used else False
    :rtype: Boolean
    """
    return SCA_TEXT in get_rhsm_status(context)


def get_available_repo_ids(context):
    """
    Retrieve repo ids of all the repositories available through the subscription-manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: Repositories that are available to the current system through the subscription-manager
    :rtype: List(string)
    """
    # Regenerated redhat.repo file ...
    # FIXME: try to come up with something less invasive than yum clean all
    # but still safe to call..
    cmd = ['yum', 'clean', 'all']
    try:
        context.call(cmd)
    except CalledProcessError as exc:
        raise StopActorExecutionError(
            'Unable to use yum successfully',
            details={'details': str(exc), 'stderr': exc.stderr}
        )

    repofiles = repofileutils.get_parsed_repofiles(context)

    # TODO: move this functionality out! Create check actor that will do
    # the inhibit. The functionality is really not good here in the current
    # shape of the leapp-repository. See the targetuserspacecreator and
    # systemfacts actor if this is moved out.
    # Issue: #486
    _inhibit_on_duplicate_repos(repofiles)
    rhsm_repos = []
    for rfile in repofiles:
        if rfile.file == _DEFAULT_RHSM_REPOFILE and rfile.data:
            rhsm_repos = [repo.repoid for repo in rfile.data]
            rhsm_repos.sort()
            break

    list_separator_fmt = '\n    - '
    if rhsm_repos:
        api.current_logger().info('The following repoids are available through RHSM:{0}{1}'
                                  .format(list_separator_fmt, list_separator_fmt.join(rhsm_repos)))
    else:
        api.current_logger().info('There are no repos available through RHSM.')
    return rhsm_repos


def _inhibit_on_duplicate_repos(repofiles):
    """
    Inhibit the upgrade if any repoid is defined multiple times.

    When that happens, it not only shows misconfigured system, but then
    we can't get details of all the available repos as well.
    """
    duplicates = repofileutils.get_duplicate_repositories(repofiles).keys()

    if not duplicates:
        return
    list_separator_fmt = '\n    - '
    api.current_logger().warning(
        'The following repoids are defined multiple times:{0}{1}'
        .format(list_separator_fmt, list_separator_fmt.join(duplicates))
    )

    reporting.create_report([
        reporting.Title('A YUM/DNF repository defined multiple times'),
        reporting.Summary(
            'The following repositories are defined multiple times:{0}{1}'
            .format(list_separator_fmt, list_separator_fmt.join(duplicates))
        ),
        reporting.Severity(reporting.Severity.MEDIUM),
        reporting.Groups([reporting.Groups.REPOSITORY]),
        reporting.Groups([reporting.Groups.INHIBITOR]),
        reporting.Remediation(hint='Remove the duplicate repository definitions.')
    ])


@with_rhsm
def get_enabled_repo_ids(context):
    """
    Retrieve repo ids of all the repositories enabled through the subscription-manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: Repositories that are enabled on the current system through the subscription-manager.
    :rtype: List(string)
    """
    with _handle_rhsm_exceptions():
        result = context.call(['subscription-manager', 'repos', '--list-enabled'], split=False)
        return _RE_REPO_UID.findall(result['stdout'])


@with_rhsm
@_rhsm_retry(max_attempts=_ATTEMPTS, sleep=_RETRY_SLEEP)
def unset_release(context):
    """
    Unset the configured release from the subscription-manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    """
    with _handle_rhsm_exceptions():
        context.call(['subscription-manager', 'release', '--unset'], split=False)


@with_rhsm
@_rhsm_retry(max_attempts=_ATTEMPTS, sleep=_RETRY_SLEEP)
def set_release(context, release):
    """
    Set the release (RHEL minor version) through the subscription-manager.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param release: Release to set the subscription-manager to.
    :type release: str
    """
    with _handle_rhsm_exceptions():
        context.call(['subscription-manager', 'release', '--set', release], split=False)


@with_rhsm
def get_release(context):
    """
    Retrieves the release the subscription-manager has been pinned to, if applicable.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: Release the subscription-manager is set to.
    :rtype: string
    """
    with _handle_rhsm_exceptions():
        result = context.call(['subscription-manager', 'release'], split=False)
        result = _RE_RELEASE.findall(result['stdout'])
        return result[0] if result else ''


@with_rhsm
@_rhsm_retry(max_attempts=_ATTEMPTS, sleep=_RETRY_SLEEP)
def refresh(context):
    """
    Calls 'subscription-manager refresh'

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    """
    with _handle_rhsm_exceptions():
        context.call(['subscription-manager', 'refresh'], split=False)


@with_rhsm
def get_existing_product_certificates(context):
    """
    Retrieves information about existing product certificates on the system.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: Paths to product certificates that are currently installed on the system.
    :rtype: List(string)
    """
    certs = []
    for path in ('/etc/pki/product', '/etc/pki/product-default'):
        if not os.path.isdir(context.full_path(path)):
            continue
        curr_certs = [os.path.join(path, f) for f in os.listdir(context.full_path(path))
                      if os.path.isfile(os.path.join(context.full_path(path), f))]
        if curr_certs:
            certs.extend(curr_certs)
    return certs


# DO NOT SET the with_rhsm decorator for this function
def set_container_mode(context):
    """
    Put RHSM into the container mode.

    Inside the container, we have to ensure the RHSM is not used AND that host
    is not affected. If the RHSM is not set into the container mode, the host
    could be affected and the generated repo file in the container could be
    affected as well (e.g. when the release is set, using rhsm, on the host).

    We want to put RHSM into the container mode always when /etc/rhsm and
    /etc/pki/entitlement directories exists, even when leapp is executed with
    --no-rhsm option. If any of these directories are missing, skip other
    actions - most likely RHSM is not installed in such a case.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    """
    if not context.is_isolated():
        api.current_logger().error('Trying to set RHSM into the container mode'
                                   'on host. Skipping the action.')
        return
    # TODO(pstodulk): check "rhsm identity" whether system is registered
    # and the container mode should be required
    if (not os.path.exists(context.full_path('/etc/rhsm'))
            or not os.path.exists(context.full_path('/etc/pki/entitlement'))):
        api.current_logger().warning(
            'Cannot set the container mode for the subscription-manager as'
            ' one of required directories is missing. Most likely RHSM is not'
            ' installed. Skipping other actions.'
        )
        return

    try:
        context.call(['ln', '-s', '/etc/rhsm', '/etc/rhsm-host'])
        context.call(['ln', '-s', '/etc/pki/entitlement', '/etc/pki/entitlement-host'])
    except CalledProcessError:
        raise StopActorExecutionError(
                message='Cannot set the container mode for the subscription-manager.')


@with_rhsm
def switch_certificate(context, rhsm_info, cert_path):
    """
    Perform all actions needed to switch the passed RHSM product certificate.

    This function will copy the certificate to /etc/pki/product, and /etc/pki/product-default if necessary, and
    remove other product certificates from there.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :param rhsm_info: An instance of the RHSMInfo model
    :type rhsm_info: RHSMInfo model
    :param cert_path: Path to the product certificate to switch to
    :type cert_path: string
    """
    for existing in rhsm_info.existing_product_certificates:
        try:
            context.remove(existing)
        except OSError:
            api.current_logger().warning('Failed to remove existing certificate: %s', existing, exc_info=True)

    for path in ('/etc/pki/product', '/etc/pki/product-default'):
        if os.path.isdir(context.full_path(path)):
            context.copy_to(cert_path, os.path.join(path, os.path.basename(cert_path)))


@with_rhsm
def scan_rhsm_info(context):
    """
    Gather all the RHSM information of the source system.

    It's not intended for gathering RHSM info about the target system within a container.

    :param context: An instance of a mounting.IsolatedActions class
    :type context: mounting.IsolatedActions class
    :return: An instance of an RHSMInfo model.
    :rtype: RHSMInfo model
    """
    info = RHSMInfo()
    info.attached_skus = get_attached_skus(context)
    info.available_repos = get_available_repo_ids(context)
    info.enabled_repos = get_enabled_repo_ids(context)
    info.release = get_release(context)
    info.existing_product_certificates.extend(get_existing_product_certificates(context))
    info.sca_detected = is_manifest_sca(context)
    return info



from __future__ import division

from leapp import reporting
from leapp.libraries.actor.checkbootavailspace import (
    check_avail_space_on_boot,
    inhibit_upgrade,
    MIN_AVAIL_BYTES_FOR_BOOT
)
from leapp.libraries.common.testutils import create_report_mocked
from leapp.utils.report import is_inhibitor


class fake_get_avail_bytes_on_boot(object):
    def __init__(self, size):
        self.size = size

    def __call__(self, *args):
        return self.size


def test_not_enough_space_available(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    # Test 0 bytes available /boot
    get_avail_bytes_on_boot = fake_get_avail_bytes_on_boot(0)
    check_avail_space_on_boot(get_avail_bytes_on_boot)

    # Test 0.1 MiB less then required in /boot
    get_avail_bytes_on_boot = fake_get_avail_bytes_on_boot(MIN_AVAIL_BYTES_FOR_BOOT - 0.1 * 2**20)
    check_avail_space_on_boot(get_avail_bytes_on_boot)

    assert reporting.create_report.called == 2


def test_enough_space_available(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    get_avail_bytes_on_boot = fake_get_avail_bytes_on_boot(MIN_AVAIL_BYTES_FOR_BOOT)
    check_avail_space_on_boot(get_avail_bytes_on_boot)

    assert reporting.create_report.called == 0


def test_inhibit_upgrade(monkeypatch):
    monkeypatch.setattr(reporting, 'create_report', create_report_mocked())

    # Test 4.2 MiB available on /boot
    bytes_available = 4.2 * 2**20
    inhibit_upgrade(bytes_available)

    assert reporting.create_report.called == 1
    assert is_inhibitor(reporting.create_report.report_fields)
    mib_needed = (MIN_AVAIL_BYTES_FOR_BOOT - bytes_available) / 2**20
    assert "needs additional {0} MiB".format(mib_needed) in reporting.create_report.report_fields['summary']



from leapp import reporting
from leapp.actors import Actor
from leapp.exceptions import StopActorExecutionError
from leapp.libraries.actor.opensshpermitrootlogincheck import global_value, semantics_changes
from leapp.libraries.common.config.version import get_source_major_version
from leapp.libraries.stdlib import api
from leapp.models import OpenSshConfig, Report
from leapp.reporting import create_report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag

COMMON_REPORT_TAGS = [
    reporting.Groups.AUTHENTICATION,
    reporting.Groups.SECURITY,
    reporting.Groups.NETWORK,
    reporting.Groups.SERVICES
]

COMMON_RESOURCES = [
    reporting.RelatedResource('package', 'openssh-server'),
    reporting.RelatedResource('file', '/etc/ssh/sshd_config')
]


class OpenSshPermitRootLoginCheck(Actor):
    """
    OpenSSH no longer allows root logins with password.

    Check the values of PermitRootLogin in OpenSSH server configuration file
    and warn about potential issues after upgrade to the next major version of RHEL.

    The RHEL8 still provided default configuration that allowed root logins,
    which can lead to possible unwanted changes during the upgrade
    """
    name = 'openssh_permit_root_login'
    consumes = (OpenSshConfig, )
    produces = (Report, )
    tags = (ChecksPhaseTag, IPUWorkflowTag, )

    def process(self):
        openssh_messages = self.consume(OpenSshConfig)
        config = next(openssh_messages, None)
        if list(openssh_messages):
            api.current_logger().warning('Unexpectedly received more than one OpenSshConfig message.')
        if not config:
            raise StopActorExecutionError(
                'Could not check openssh configuration', details={'details': 'No OpenSshConfig facts found.'}
            )

        if get_source_major_version() == '7':
            self.process7to8(config)
        elif get_source_major_version() == '8':
            self.process8to9(config)
        else:
            api.current_logger().warning('Unknown source major version: {} (expecting 7 or 8)'
                                         .format(get_source_major_version()))

    def process7to8(self, config):
        # when the config was not modified, we can pass this check and let the
        # rpm handle the configuration file update
        if not config.modified:
            return

        # When the configuration does not contain *any* PermitRootLogin directive and
        # the configuration file was locally modified, it will not get updated by
        # RPM and the user might be locked away from the server with new default
        if not config.permit_root_login:
            create_report([
                reporting.Title('Possible problems with remote login using root account'),
                reporting.Summary(
                    'OpenSSH configuration file does not explicitly state '
                    'the option PermitRootLogin in sshd_config file, '
                    'which will default in RHEL8 to "prohibit-password".'
                ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups(COMMON_REPORT_TAGS),
                reporting.Remediation(
                    hint='If you depend on remote root logins using passwords, consider '
                         'setting up a different user for remote administration or adding '
                         '"PermitRootLogin yes" to sshd_config. '
                         'If this change is ok for you, add explicit '
                         '"PermitRootLogin prohibit-password" to your sshd_config '
                         'to ignore this inhibitor'
                ),
                reporting.Groups([reporting.Groups.INHIBITOR])
            ] + COMMON_RESOURCES)
            return

        # Check if there is at least one PermitRootLogin other than "no"
        # in match blocks (other than Match All).
        # This usually means some more complicated setup depending on the
        # default value being globally "yes" and being overwritten by this
        # match block
        if semantics_changes(config):
            create_report([
                reporting.Title('OpenSSH configured to allow root login'),
                reporting.Summary(
                    'OpenSSH is configured to deny root logins in match '
                    'blocks, but not explicitly enabled in global or '
                    '"Match all" context. This update changes the '
                    'default to disable root logins using passwords '
                    'so your server might get inaccessible.'
                ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups(COMMON_REPORT_TAGS),
                reporting.Remediation(
                    hint='Consider using different user for administrative '
                         'logins or make sure your configuration file '
                         'contains the line "PermitRootLogin yes" '
                         'in global context if desired.'
                ),
                reporting.Groups([reporting.Groups.INHIBITOR])
            ] + COMMON_RESOURCES)

    def process8to9(self, config):
        # RHEL8 default sshd configuration file is not modified: It will get replaced by rpm and
        # root will no longer be able to connect through ssh. This will probably result in many
        # false positives so it will have to be waived a lot
        if not config.modified:
            create_report([
                reporting.Title('Possible problems with remote login using root account'),
                reporting.Summary(
                    'OpenSSH configuration file will get updated to RHEL9 '
                    'version, no longer allowing root login with password. '
                    'It is a good practice to use non-root administrative '
                    'user and non-password authentications, but if you rely '
                    'on the remote root login, this change can lock you out '
                    'of this system.'
                ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups(COMMON_REPORT_TAGS),
                reporting.Remediation(
                    hint='If you depend on remote root logins using passwords, '
                         'consider setting up a different user for remote '
                         'administration or adding a comment into the '
                         'sshd_config next to the "PermitRootLogin yes" directive '
                         'to prevent rpm replacing it during the upgrade.'
                ),
                reporting.Groups([reporting.Groups.INHIBITOR])
            ] + COMMON_RESOURCES)
        # If the configuration is modified and contains any directive allowing
        # root login (which is in default configuration), we are upgrading to
        # RHEL9 keeping the old "security policy", which might keep the root
        # login unexpectedly open. This might be just high priority warning
        if global_value(config, 'prohibit-password') == 'yes':
            create_report([
                reporting.Title('Remote root logins globally allowed using password'),
                reporting.Summary(
                    'RHEL9 no longer allows remote root logins, but the '
                    'server configuration explicitly overrides this default. '
                    'The configuration file will not be updated and root is '
                    'still going to be allowed to login with password. '
                    'This is not recommended and considered as a security risk.'
                ),
                reporting.Severity(reporting.Severity.HIGH),
                reporting.Groups(COMMON_REPORT_TAGS),
                reporting.Remediation(
                    hint='If you depend on remote root logins using passwords, '
                         'consider setting up a different user for remote '
                         'administration. Otherwise you can ignore this message.'
                )
            ] + COMMON_RESOURCES)



import pytest

from leapp.libraries.actor.cupsscanner import environment_setup_check

testdata = (
    ('\n', False),
    ('Something else\n', False),
    ('#PassEnv smth\n', False),
    ('   #SetEnv smth\n', False),
    ('PassEnv smth\n', True),
    ('SetEnv smth\n', True),
    ('PassEnv\n', False),
    ('SetEnv\n', False),
    ('PassEnv smth\nSetEnv smth\n', True)
)


class MockCUPSD(object):
    def __init__(self, content):
        self.content = content

    def read(self, path):
        if path:
            return self.content.splitlines(True)
        return None


@pytest.mark.parametrize("content, expected", testdata)
def test_environment_setup_check(content, expected):
    config = MockCUPSD(content)

    ret = environment_setup_check('does_not_matter', config.read)

    assert ret == expected



from leapp.actors import Actor
from leapp.libraries.actor.checkntp import check_ntp
from leapp.models import DistributionSignedRPM, NtpMigrationDecision, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckNtp(Actor):
    """
    Check if ntp and/or ntpdate configuration needs to be migrated.
    """

    name = 'check_ntp'
    consumes = (DistributionSignedRPM,)
    produces = (Report, NtpMigrationDecision)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        installed_packages = set()

        signed_rpms = self.consume(DistributionSignedRPM)
        for rpm_pkgs in signed_rpms:
            for pkg in rpm_pkgs.items:
                installed_packages.add(pkg.name)

        self.produce(check_ntp(installed_packages))



from leapp.actors import Actor
from leapp.libraries.common.rpms import has_package
from leapp.libraries.stdlib import api
from leapp.models import CopyFile, DistributionSignedRPM, TargetUserSpacePreupgradeTasks
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag

PLUGIN_PKGNAME = "kpatch-dnf"
CONFIG_PATH = "/etc/dnf/plugins/kpatch.conf"


class CheckKpatch(Actor):
    """
    Carry over kpatch-dnf and it's config into the container

    Check is kpatch-dnf plugin is installed and if it is, install it and copy
    over the config file so that the plugin can make a decision on whether any
    kpatch-patch packages need to be installed during in-place upgrade.
    """

    name = 'check_kpatch'
    consumes = (DistributionSignedRPM,)
    produces = (TargetUserSpacePreupgradeTasks,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        if has_package(DistributionSignedRPM, PLUGIN_PKGNAME):
            api.produce(TargetUserSpacePreupgradeTasks(
                install_rpms=[PLUGIN_PKGNAME],
                copy_files=[CopyFile(src=CONFIG_PATH)]))



import pytest

from leapp.libraries.actor.cupsfiltersmigrate import BROWSED_CONFIG, update_cups_browsed

testdata = [
    {'cups-filters': '/etc/cups/cups-browsed.conf'},
    {'cups-filters': ''},
    {'ble': ''}
]


class MockLogger(object):
    def __init__(self):
        self.debugmsg = ''
        self.errmsg = ''

    def debug(self, message):
        self.debugmsg += message

    def error(self, message):
        self.errmsg += message


class MockPackage(object):
    def __init__(self, name, config):
        self.name = name
        self.config = config
        self.config_content = ''


class MockPackageSet(object):
    def __init__(self):
        self.installed_packages = None

    def add_packages(self, pkgs):
        if self.installed_packages is None:
            self.installed_packages = []

        for rpm, config in pkgs.items():
            self.installed_packages.append(MockPackage(rpm, config))

    def is_installed(self, pkg):
        for rpm in self.installed_packages:
            if pkg == rpm.name:
                return True
        return False

    def append_content(self, path, content):
        found = False

        for rpm in self.installed_packages:
            if path == rpm.config:
                found = True
                rpm.config_content += content
        if not found:
            raise IOError('Error during writing to file: {}.'.format(path))

    def check_content(self, path, content):
        found = False

        for rpm in self.installed_packages:
            if path == rpm.config and content in rpm.config_content:
                found = True

        return found


class ExpectedOutput(object):
    def __init__(self):
        self.debugmsg = ''
        self.errmsg = ''

    def create(self, rpms):
        error_list = []

        for pkg, config in rpms.items():
            if pkg == 'cups-filters':
                self.debugmsg += 'Updating cups-browsed configuration file {}.'.format(BROWSED_CONFIG)
                if config == '':
                    error_list.append((BROWSED_CONFIG, 'Error during '
                                       'writing to file: {}.'.format(BROWSED_CONFIG)))

        if error_list:
            self.errmsg = ('The files below have not been modified '
                           '(error message included):' +
                           ''.join(['\n    - {}: {}'.format(err[0], err[1])
                                   for err in error_list]))


@pytest.mark.parametrize("rpms", testdata)
def test_update_cups_browsed(rpms):
    logger = MockLogger()
    installed_packages = MockPackageSet()

    installed_packages.add_packages(rpms)

    expected = ExpectedOutput()
    expected.create(rpms)

    update_cups_browsed(logger.debug,
                        logger.error,
                        installed_packages.is_installed,
                        installed_packages.append_content,
                        installed_packages.check_content)

    assert expected.debugmsg == logger.debugmsg
    assert expected.errmsg == logger.errmsg



from leapp.actors import Actor
from leapp.libraries.actor import checkdeprecatedrpmsignature
from leapp.models import CryptoPolicyInfo, InstalledRPM, Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckDeprecatedRPMSignature(Actor):
    """
    Check whether any packages signed by RSA/SHA1 are installed

    Crypto policies on RHEL 9 disallow use of the SHA-1 hash algorithm.
    Regarding the IPU, the major impact is around RPMs with RSA/SHA1 signatures
    that cannot be handled during the upgrade when SHA1 is not explicitly
    allowed (e.g. setting the LEGACY mode; which has serious impacts).

    The problem is that when rpm tries to verify the bad signature of the
    installed rpm, it ends with error regarding the new openssl policies
    (same for dnf). In such a case, rpm nor dnf prints the name of the package,
    nor report anything else useful. In case of the IPU this happens when
    the bad packages is supposed to be removed (includes upgrade, downgrade,...)
    from the system during the dnf transaction. There is no way how we could
    handle this situation from our DNF plugin (unless we disable crypto in DNF
    completely, which is not desired). Also, in case the key has not been
    imported previously, the issue will not be seen as there is no way how to
    check the signature.

    The current hotfix implementation inhibits the upgrade when any RPMs with
    RSA/SHA1 are installed on the system and:
      * the system wide crypto policy is not LEGACY or DEFAULT:SHA1
    In other cases when such RPMs are installed on the system, just report the
    high risk informing users about the situation.
    """

    name = 'check_deprecated_rpm_signature'
    consumes = (CryptoPolicyInfo, InstalledRPM)
    produces = (Report,)
    tags = (IPUWorkflowTag, ChecksPhaseTag)

    def process(self):
        checkdeprecatedrpmsignature.process()



import contextlib
import os
from functools import partial

import pytest

from leapp.libraries.actor import scan_target_os_iso
from leapp.libraries.common.mounting import MountError
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api, CalledProcessError
from leapp.models import CustomTargetRepository, TargetOSInstallationImage


def fail_if_called(fail_reason, *args, **kwargs):
    assert False, fail_reason


def test_determine_rhel_version_determination_unexpected_iso_structure_or_invalid_mountpoint(monkeypatch):
    iso_mountpoint = '/some/mountpoint'

    run_mocked = partial(fail_if_called,
                         'No commands should be called when mounted ISO mountpoint has unexpected structure.')
    monkeypatch.setattr(scan_target_os_iso, 'run', run_mocked)

    def isdir_mocked(path):
        assert path == '/some/mountpoint/BaseOS/Packages', 'Only the contents of BaseOS/Packages should be examined.'
        return False

    monkeypatch.setattr(os.path, 'isdir', isdir_mocked)

    determined_version = scan_target_os_iso.determine_rhel_version_from_iso_mountpoint(iso_mountpoint)
    assert not determined_version


def test_determine_rhel_version_valid_iso(monkeypatch):
    iso_mountpoint = '/some/mountpoint'

    def isdir_mocked(path):
        return True

    def listdir_mocked(path):
        assert path == '/some/mountpoint/BaseOS/Packages', 'Only the contents of BaseOS/Packages should be examined.'
        return ['xz-5.2.4-4.el8_6.x86_64.rpm',
                'libmodman-2.0.1-17.el8.i686.rpm',
                'redhat-release-8.7-0.3.el8.x86_64.rpm',
                'redhat-release-eula-8.7-0.3.el8.x86_64.rpm']

    def run_mocked(cmd, *args, **kwargs):
        rpm2cpio_output = 'rpm2cpio_output'
        if cmd[0] == 'rpm2cpio':
            assert cmd == ['rpm2cpio', '/some/mountpoint/BaseOS/Packages/redhat-release-8.7-0.3.el8.x86_64.rpm']
            return {'stdout': rpm2cpio_output}
        if cmd[0] == 'cpio':
            assert cmd == ['cpio', '--extract', '--to-stdout', './etc/redhat-release']
            assert kwargs['stdin'] == rpm2cpio_output
            return {'stdout': 'Red Hat Enterprise Linux Server release 7.9 (Maipo)'}
        raise ValueError('Unexpected command has been called.')

    monkeypatch.setattr(os.path, 'isdir', isdir_mocked)
    monkeypatch.setattr(os, 'listdir', listdir_mocked)
    monkeypatch.setattr(scan_target_os_iso, 'run', run_mocked)

    determined_version = scan_target_os_iso.determine_rhel_version_from_iso_mountpoint(iso_mountpoint)
    assert determined_version == '7.9'


def test_determine_rhel_version_valid_iso_no_rh_release(monkeypatch):
    iso_mountpoint = '/some/mountpoint'

    def isdir_mocked(path):
        return True

    def listdir_mocked(path):
        assert path == '/some/mountpoint/BaseOS/Packages', 'Only the contents of BaseOS/Packages should be examined.'
        return ['xz-5.2.4-4.el8_6.x86_64.rpm',
                'libmodman-2.0.1-17.el8.i686.rpm',
                'redhat-release-eula-8.7-0.3.el8.x86_64.rpm']

    run_mocked = partial(fail_if_called, 'No command should be called if the redhat-release package is not present.')

    monkeypatch.setattr(os.path, 'isdir', isdir_mocked)
    monkeypatch.setattr(os, 'listdir', listdir_mocked)
    monkeypatch.setattr(scan_target_os_iso, 'run', run_mocked)

    determined_version = scan_target_os_iso.determine_rhel_version_from_iso_mountpoint(iso_mountpoint)
    assert determined_version == ''


def test_determine_rhel_version_rpm_extract_fails(monkeypatch):
    iso_mountpoint = '/some/mountpoint'

    def isdir_mocked(path):
        return True

    def listdir_mocked(path):
        assert path == '/some/mountpoint/BaseOS/Packages', 'Only the contents of BaseOS/Packages should be examined.'
        return ['redhat-release-8.7-0.3.el8.x86_64.rpm']

    def run_mocked(cmd, *args, **kwargs):
        raise CalledProcessError(message='Ooops.', command=cmd, result=2)

    monkeypatch.setattr(os.path, 'isdir', isdir_mocked)
    monkeypatch.setattr(os, 'listdir', listdir_mocked)
    monkeypatch.setattr(scan_target_os_iso, 'run', run_mocked)

    determined_version = scan_target_os_iso.determine_rhel_version_from_iso_mountpoint(iso_mountpoint)
    assert determined_version == ''


@pytest.mark.parametrize('etc_rh_release_contents', ('',
                                                     'Red Hat Enterprise Linux Server',
                                                     'Fedora release 35 (Thirty Five)'))
def test_determine_rhel_version_unexpected_etc_rh_release_contents(monkeypatch, etc_rh_release_contents):
    iso_mountpoint = '/some/mountpoint'

    def isdir_mocked(path):
        return True

    def listdir_mocked(path):
        assert path == '/some/mountpoint/BaseOS/Packages', 'Only the contents of BaseOS/Packages should be examined.'
        return ['redhat-release-8.7-0.3.el8.x86_64.rpm']

    def run_mocked(cmd, *args, **kwargs):
        if cmd[0] == 'rpm2cpio':
            return {'stdout': 'rpm2cpio_output'}
        if cmd[0] == 'cpio':
            return {'stdout': etc_rh_release_contents}
        raise ValueError('Actor called an unexpected command: {0}'.format(cmd))

    monkeypatch.setattr(os.path, 'isdir', isdir_mocked)
    monkeypatch.setattr(os, 'listdir', listdir_mocked)
    monkeypatch.setattr(scan_target_os_iso, 'run', run_mocked)

    determined_version = scan_target_os_iso.determine_rhel_version_from_iso_mountpoint(iso_mountpoint)
    assert determined_version == ''


@pytest.mark.parametrize('iso_envar_set', (True, False))
def test_iso_detection_with_no_iso(monkeypatch, iso_envar_set):
    envars = {'LEAPP_TARGET_ISO': '/target_iso'} if iso_envar_set else {}
    mocked_actor = CurrentActorMocked(envars=envars)
    monkeypatch.setattr(api, 'current_actor', mocked_actor)
    monkeypatch.setattr(api, 'produce', produce_mocked())

    scan_target_os_iso.inform_ipu_about_request_to_use_target_iso()
    assert bool(api.produce.called) == iso_envar_set


def test_iso_mounting_failed(monkeypatch):
    envars = {'LEAPP_TARGET_ISO': '/target_iso'}
    mocked_actor = CurrentActorMocked(envars=envars)
    monkeypatch.setattr(api, 'current_actor', mocked_actor)
    monkeypatch.setattr(api, 'produce', produce_mocked())

    def raise_mount_error_when_called():
        raise MountError('MountError')

    monkeypatch.setattr(scan_target_os_iso, 'LoopMount', raise_mount_error_when_called)

    scan_target_os_iso.inform_ipu_about_request_to_use_target_iso()
    assert api.produce.called

    assert len(api.produce.model_instances) == 1
    assert not api.produce.model_instances[0].was_mounted_successfully


@pytest.mark.parametrize(('repodirs_in_iso', 'expected_repoids'),
                         (((), ()),
                          (('BaseOS',), ('BaseOS',)),
                          (('BaseOS', 'AppStream'), ('BaseOS', 'AppStream')),
                          (('BaseOS', 'AppStream', 'UnknownRepo'), ('BaseOS', 'AppStream'))))
def test_iso_repository_detection(monkeypatch, repodirs_in_iso, expected_repoids):
    iso_path = '/target_iso'
    envars = {'LEAPP_TARGET_ISO': iso_path}
    mocked_actor = CurrentActorMocked(envars=envars)

    @contextlib.contextmanager
    def always_successful_loop_mount(*args, **kwargs):
        yield

    def mocked_os_path_exits(path):
        if path == iso_path:
            return True
        raise ValueError('Only the ISO path should be probed for existence.')

    def mocked_os_listdir(path):
        # Add some extra files as an ISO will always have some extra files in / as the ones parametrizing this test
        return list(repodirs_in_iso + ('eula.txt', 'grub', 'imgs'))

    monkeypatch.setattr(api, 'current_actor', mocked_actor)
    monkeypatch.setattr(api, 'produce', produce_mocked())
    monkeypatch.setattr(scan_target_os_iso, 'LoopMount', always_successful_loop_mount)
    monkeypatch.setattr(os.path, 'exists', mocked_os_path_exits)
    monkeypatch.setattr(os, 'listdir', mocked_os_listdir)
    monkeypatch.setattr(scan_target_os_iso, 'determine_rhel_version_from_iso_mountpoint', lambda iso_mountpoint: '7.9')

    scan_target_os_iso.inform_ipu_about_request_to_use_target_iso()

    produced_msgs = api.produce.model_instances
    assert len(produced_msgs) == 1 + len(expected_repoids)

    produced_custom_repo_msgs = []
    target_iso_msg = None
    for produced_msg in produced_msgs:
        if isinstance(produced_msg, CustomTargetRepository):
            produced_custom_repo_msgs.append(produced_msg)
        else:
            assert not target_iso_msg, 'Actor is expected to produce only one TargetOSInstallationImage msg'
            target_iso = produced_msg

    # Do not explicitly instantiate model instances of what we expect the model instance to look like. Instead check
    # for expected structural properties, leaving the actor implementation flexibility (e.g. choice of the mountpoint)
    iso_mountpoint = target_iso.mountpoint

    assert target_iso.was_mounted_successfully
    assert target_iso.rhel_version == '7.9'

    expected_repos = {(repoid, 'file://' + os.path.join(iso_mountpoint, repoid)) for repoid in expected_repoids}
    actual_repos = {(repo.repoid, repo.baseurl) for repo in produced_custom_repo_msgs}
    assert expected_repos == actual_repos



from leapp.actors import Actor
from leapp.libraries.actor.checkipaserver import ipa_inhibit_upgrade, ipa_warn_pkg_installed
from leapp.models import IpaInfo
from leapp.reporting import Report
from leapp.tags import ChecksPhaseTag, IPUWorkflowTag


class CheckIPAServer(Actor):
    """
    Check for ipa-server and inhibit upgrade
    """

    name = "check_ipa_server"
    consumes = (IpaInfo,)
    produces = (Report,)
    tags = (ChecksPhaseTag, IPUWorkflowTag)

    def process(self):
        for ipainfo in self.consume(IpaInfo):
            if ipainfo.is_server_configured:
                self.log.info(
                    "IdM server instance detected, inhibit upgrade"
                )
                ipa_inhibit_upgrade(ipainfo)
            elif ipainfo.has_server_package:
                self.log.info("Unused ipa-server package detected")
                ipa_warn_pkg_installed(ipainfo)



from leapp import reporting
from leapp.libraries.stdlib import api
from leapp.models import DialogModel


def check_dialogs(inhibit_if_no_userchoice=True):
    results = list(api.consume(DialogModel))
    for dialog in results:
        sections = dialog.answerfile_sections
        summary = ('One or more sections in answerfile are missing user choices: {}\n'
                   'For more information consult https://red.ht/leapp-dialogs.')
        dialog_resources = [reporting.RelatedResource('dialog', s) for s in sections]
        dialogs_remediation = ('Please register user choices with leapp answer cli command or by manually editing '
                               'the answerfile.')
        # FIXME: Enable more choices once we can do multi-command remediations
        cmd_remediation = [['leapp', 'answer', '--section', "{}={}".format(s, choice)]
                           for s, choices in dialog.answerfile_sections.items() for choice in choices[:1]]
        report_data = [reporting.Title('Missing required answers in the answer file'),
                       reporting.Severity(reporting.Severity.HIGH),
                       reporting.Summary(summary.format('\n'.join(sections))),
                       reporting.Groups([reporting.Groups.INHIBITOR] if inhibit_if_no_userchoice else []),
                       reporting.Remediation(hint=dialogs_remediation, commands=cmd_remediation),
                       reporting.Key(dialog.key)]
        reporting.create_report(report_data + dialog_resources)



from leapp.actors import Actor
from leapp.libraries.common import grub
from leapp.libraries.common.config import architecture
from leapp.models import GrubInfo
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class ScanGrubDeviceName(Actor):
    """
    Find the name of the block devices where GRUB is located
    """

    name = 'scan_grub_device_name'
    consumes = ()
    produces = (GrubInfo,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        if architecture.matches_architecture(architecture.ARCH_S390X):
            return

        devices = grub.get_grub_devices()
        grub_info = GrubInfo(orig_devices=devices)
        grub_info.orig_device_name = devices[0] if len(devices) == 1 else None
        self.produce(grub_info)



from leapp.actors import Actor
from leapp.libraries.actor import vsftpdconfigupdate
from leapp.models import VsftpdFacts
from leapp.tags import ApplicationsPhaseTag, IPUWorkflowTag


class VsftpdConfigUpdate(Actor):
    """
    Modifies vsftpd configuration files on the target RHEL-8 system so that the effective
    configuration is the same, where possible. This means doing two things:
    1. Reverting the default configuration file (/etc/vsftpd/vsftpd.conf) to its state
       before the upgrade (where it makes sense), if the configuration file was being used
       with its default content (i.e., unmodified) on the source system (the configuration
       file gets replaced with a new version during the RPM upgrade in this case).
       The anonymous_enable option falls in this category.
    2. Adding 'option=old_effective_value' to configuration files for options whose default
       value has changed, if the option is not explicitly specified in the configuration file.
       The strict_ssl_read_eof option falls in this category.
    3. Disabling options that cannot be enabled, otherwise vsftpd wouldn't work.
       The tcp_wrappers option falls in this category.
    """

    name = 'vsftpd_config_update'
    consumes = (VsftpdFacts,)
    produces = ()
    tags = (ApplicationsPhaseTag, IPUWorkflowTag)

    def process(self):
        try:
            vsftpd_facts = next(self.consume(VsftpdFacts))
        except StopIteration:
            return
        vsftpdconfigupdate.migrate_configs(vsftpd_facts)



import logging
from functools import partial

import pytest

from leapp.libraries.actor import sctpdlm, sctplib
from leapp.libraries.common.testutils import CurrentActorMocked
from leapp.models import ActiveKernelModule, ActiveKernelModulesFacts

FILENAME_SCTP = 'sctp'
FILENAME_NO_SCTP = 'no_sctp'
SRC_VER = '7.6'

logger = logging.getLogger(__name__)


def test_anyfile(tmpdir):
    file1 = tmpdir.join('file1')
    file2 = tmpdir.join('file2')
    file1.write('I am not empty')
    file2.write('And me either')

    assert sctplib.anyfile((str(file1),))
    assert sctplib.anyfile((str(file1), str(tmpdir)))
    assert not sctplib.anyfile((str(tmpdir),))
    assert not sctplib.anyfile(('Iam not exist',))


def test_is_module_loaded(monkeypatch):
    monkeypatch.setattr(
        sctplib.api,
        'current_actor',
        CurrentActorMocked(
            src_ver=SRC_VER,
            msgs=[
                ActiveKernelModulesFacts(
                    kernel_modules=[
                        ActiveKernelModule(
                            filename=FILENAME_SCTP, parameters=()
                        ),
                    ]
                ),
            ],
        ),
    )
    assert sctplib.is_module_loaded(FILENAME_SCTP)
    assert not sctplib.is_module_loaded('not exists filename')


@pytest.mark.parametrize(
    (
        'actor',
        'exp_return',
        'anyfile_returns',
        'check_dlm_cfgfile_returns',
        'check_dlm_sysconfig_returns',
        'text_in_log',
    ),
    [
        # test if module name is sctp
        (
            CurrentActorMocked(
                src_ver=SRC_VER,
                msgs=[
                    ActiveKernelModulesFacts(
                        kernel_modules=[
                            ActiveKernelModule(
                                filename=FILENAME_SCTP, parameters=()
                            )
                        ]
                    )
                ],
            ),
            True,
            False,
            False,
            False,
            '',
        ),
        # test if module name is different, but one of lksctp is present
        (
            CurrentActorMocked(
                src_ver=SRC_VER,
                msgs=[
                    ActiveKernelModulesFacts(
                        kernel_modules=[
                            ActiveKernelModule(
                                filename=FILENAME_NO_SCTP, parameters=()
                            )
                        ]
                    )
                ],
            ),
            True,
            True,
            False,
            False,
            'lksctp files',
        ),
        # test if check_dlm_cfgfile is True
        (
            CurrentActorMocked(
                src_ver=SRC_VER,
                msgs=[
                    ActiveKernelModulesFacts(
                        kernel_modules=[
                            ActiveKernelModule(
                                filename=FILENAME_NO_SCTP, parameters=()
                            )
                        ]
                    )
                ],
            ),
            True,
            False,
            True,
            False,
            'dlm.conf',
        ),
        # test if check_dlm_sysconfig is True
        (
            CurrentActorMocked(
                src_ver=SRC_VER,
                msgs=[
                    ActiveKernelModulesFacts(
                        kernel_modules=[
                            ActiveKernelModule(
                                filename=FILENAME_NO_SCTP, parameters=()
                            )
                        ]
                    )
                ],
            ),
            True,
            False,
            False,
            True,
            'sysconfig',
        ),
    ],
)
def test_is_sctp_used(
    actor,
    exp_return,
    anyfile_returns,
    check_dlm_cfgfile_returns,
    check_dlm_sysconfig_returns,
    text_in_log,
    monkeypatch,
    caplog,
):
    monkeypatch.setattr(sctplib.api, 'current_actor', actor)
    monkeypatch.setattr(sctplib, 'anyfile', lambda arg: anyfile_returns)
    monkeypatch.setattr(
        sctpdlm, 'check_dlm_cfgfile', lambda: check_dlm_cfgfile_returns
    )
    monkeypatch.setattr(
        sctpdlm, 'check_dlm_sysconfig', lambda: check_dlm_sysconfig_returns
    )
    with caplog.at_level(logging.DEBUG):
        assert sctplib.is_sctp_used() == exp_return
    if text_in_log:
        assert text_in_log in caplog.text


class RunMocked(object):
    """Simple mock class for leapp.libraries.stdlib.run."""

    def __init__(self, exc_type=None):
        """if exc_type provided, then it will be raised on
        instance call.

        :type exc_type: None or BaseException
        """
        self.exc_type = exc_type

    def __call__(self, *args, **kwargs):
        if self.exc_type:
            logger.info('Mocked `run` raising %r', self.exc_type)
            raise self.exc_type()
        logger.info('Mocked `run` passed without exp.')


@pytest.mark.parametrize(
    ('run_fails', 'exp_return', 'text_in_log'),
    [
        (True, False, 'Nothing regarding SCTP was found on journal.'),
        (False, True, 'Found logs regarding SCTP on journal.'),
    ],
)
def test_was_sctp_used(
    monkeypatch, caplog, run_fails, exp_return, text_in_log
):
    monkeypatch.setattr(
        sctplib,
        'run',
        RunMocked(
            exc_type=partial(
                sctplib.CalledProcessError, 'message', 'command', 'result'
            )
            if run_fails
            else None
        ),
    )
    with caplog.at_level(logging.DEBUG):
        assert sctplib.was_sctp_used() == exp_return
    if text_in_log:
        assert text_in_log in caplog.text


@pytest.mark.parametrize(
    (
        'is_sctp_used_returns',
        'was_sctp_used_returns',
        'exp_return',
        'text_in_log',
    ),
    [
        (True, False, True, 'SCTP is being used.'),
        (False, True, True, 'SCTP was used.'),
        (False, False, False, 'SCTP is not being used and neither wanted.'),
    ],
)
def test_is_sctp_wanted(
    is_sctp_used_returns,
    was_sctp_used_returns,
    exp_return,
    text_in_log,
    monkeypatch,
    caplog,
):
    monkeypatch.setattr(sctplib, 'is_sctp_used', lambda: is_sctp_used_returns)
    monkeypatch.setattr(
        sctplib, 'was_sctp_used', lambda: was_sctp_used_returns
    )
    with caplog.at_level(logging.DEBUG):
        assert sctplib.is_sctp_wanted() == exp_return
    if text_in_log:
        assert text_in_log in caplog.text



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class SatellitePostgresqlFacts(Model):
    topic = SystemFactsTopic

    local_postgresql = fields.Boolean(default=True)
    """ Whether or not PostgreSQL is installed on the same system """
    old_var_lib_pgsql_data = fields.Boolean(default=False)
    """ Whether or not there is old PostgreSQL data in /var/lib/pgsql/data """
    same_partition = fields.Boolean(default=True)
    """ Whether or not target and source postgresql data will stay on the same partition """
    space_required = fields.Nullable(fields.Integer())
    """ How many bytes are required on the target partition """
    space_available = fields.Nullable(fields.Integer())
    """ How many bytes are available on the target partition """


class SatelliteFacts(Model):
    topic = SystemFactsTopic

    has_foreman = fields.Boolean(default=False)
    """Whether or not foreman is installed on this system"""
    has_katello_installer = fields.Boolean(default=True)
    """Whether or not the installer supports Katello additions"""
    postgresql = fields.Model(SatellitePostgresqlFacts)
    """ Foreman related PostgreSQL facts """



from leapp.models import SSSDConfig, SSSDDomainConfig


class SSSDFactsLibrary(object):
    """
    Helper library from SSSDFacts actor to allow unit testing.
    """
    def __init__(self, config):
        self.config = config

    def process(self):
        """
        Check SSSD configuration for changes in RHEL8 and return them in SSSDConfig
        model.

        These changes are:
        - id_provider=local is no longer supported and will be ignored
        - ldap_groups_use_matching_rule_in_chain was removed and will be ignored
        - ldap_initgroups_use_matching_rule_in_chain was removed and will be ignored
        - ldap_sudo_include_regexp changed default from true to false
        """
        facts = SSSDConfig(domains=[])

        for section in self.config.sections():
            # We are interested only in domains.
            if not section.startswith('domain/'):
                continue

            domain = SSSDDomainConfig(
                name=section[len('domain/'):],
                options=[]
            )

            steps = {
                'local_provider': self.checkLocalProvider(domain.name),
                'groups_chain': self.checkRemovedOption(
                    domain.name, 'ldap_groups_use_matching_rule_in_chain'
                ),
                'initgroups_chain': self.checkRemovedOption(
                    domain.name, 'ldap_initgroups_use_matching_rule_in_chain'
                ),
                'sudo_regexp': self.checkSudoRegexp(domain.name)
            }

            for key, value in steps.items():
                if value:
                    domain.options.append(key)

            facts.domains.append(domain)

        return facts

    def checkLocalProvider(self, domain):
        """
        Check if any id_provider=local is configured.
        """
        provider = self.get_provider(domain, 'id_provider')
        if provider != 'local':
            return False

        return True

    def checkRemovedOption(self, domain, option):
        """
        Check if specific option that has been removed is present.
        """
        section = self.get_domain_section(domain)
        if not self.config.has_option(section, option):
            return False

        return True

    def checkSudoRegexp(self, domain):
        """
        Check if ldap_sudo_include_regexp is not set explicitly.
        """
        section = self.get_domain_section(domain)
        provider = self.get_provider(domain, 'sudo_provider', ['id_provider'])

        if provider not in ['ad', 'ldap']:
            return False

        if self.config.has_option(section, 'ldap_sudo_include_regexp'):
            return False

        return True

    def get_domain_section(self, domain):
        """
        Convert domain name to ini format section.
        """
        return 'domain/' + domain

    def get_provider(self, domain, provider, fallbacks=None):
        """
        Return configured provider.
        """
        providers = [provider] + (fallbacks if fallbacks else [])

        section = self.get_domain_section(domain)
        for option in providers:
            if self.config.has_option(section, option):
                return self.config.get(section, option)

        return None



from leapp import reporting
from leapp.libraries.common import grub
from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api
from leapp.models import DefaultGrubInfo, FirmwareFacts, GrubCfgBios

URL = (
    'https://www.ibm.com/docs/en/linux-on-systems?topic=lpo-linux-distributions-virtualization'
    '-options-power8-power9-linux-power-systems'
)
TITLE = 'Linux distributions and virtualization options for POWER8 and POWER9 Linux on Power systems'


def process():
    default_grub_msg = next(api.consume(DefaultGrubInfo), None)
    grub_cfg = next(api.consume(GrubCfgBios), None)
    ff = next(api.consume(FirmwareFacts), None)
    if None in (default_grub_msg, grub_cfg):
        api.current_logger().debug(
            'Skipping execution. No DefaultGrubInfo and GrubCfgBios messages have '
            'been produced'
        )
        return

    if not architecture.matches_architecture(architecture.ARCH_PPC64LE):
        return

    if ff and ff.ppc64le_opal:
        reporting.create_report([
            reporting.Title(
                'Leapp cannot continue with upgrade on "ppc64le" bare metal systems'
            ),
            reporting.Summary(
                'In-place upgrade to RHEL 9 is not supported on POWER8 and POWER9 bare metal systems. '
                'For more information, refer to the following article: {}'.format(URL)
            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups(['inhibitor']),
            reporting.Groups([reporting.Groups.BOOT]),
            reporting.ExternalLink(
                url=URL,
                title=TITLE
            )
        ])

    if (
        not grub_cfg.insmod_bls and grub.is_blscfg_enabled_in_defaultgrub(default_grub_msg)
    ):
        reporting.create_report([
            reporting.Title(
                'Leapp will execute "grub2-mkconfig" to fix BLS Grub configuration.'
            ),
            reporting.Summary(
                'On "ppc64le" systems with BLS enabled, the GRUB configuration is not '
                'properly converted after the upgrade and Leapp has to run "grub2-mkconfig" '
                '-o /boot/grub2/grub.cfg command in order to fix an issue with booting into '
                'the RHEL 8 kernel instead of RHEL 9.'

            ),
            reporting.Severity(reporting.Severity.HIGH),
            reporting.Groups([reporting.Groups.BOOT]),
        ])



from leapp.models import fields, Model
from leapp.topics import SystemFactsTopic


class RemovedPAMModules(Model):
    """
    PAM modules that were removed from RHEL8 but are in current configuration.
    """
    topic = SystemFactsTopic

    modules = fields.List(fields.String())
    """
    List of PAM modules that were detected in current configuration but
    are no longer available in RHEL8.
    """



import os

import pytest

from leapp.libraries.actor import migrateopensslconf
from leapp.libraries.common.testutils import CurrentActorMocked, logger_mocked
from leapp.libraries.stdlib import CalledProcessError


class PathExistsMocked(object):
    def __init__(self, existing_files=None):
        self.called = 0
        self._existing_files = existing_files if existing_files else []

    def __call__(self, fpath):
        self.called += 1
        return fpath in self._existing_files


class IsOpensslModifiedMocked(object):
    def __init__(self, ret_values):
        self._ret_values = ret_values
        # ret_values is list of bools to return on each call. ret_values.pop(0)
        # if the list becomes empty, returns False

        self.called = 0

    def __call__(self):
        self.called += 1
        if not self._ret_values:
            return False
        return self._ret_values.pop(0)


class SafeMVFileMocked(object):
    def __init__(self, ret_values):
        self._ret_values = ret_values
        # ret_values is list of bools to return on each call. ret_values.pop(0)
        # if the list becomes empty, returns False

        self.called = 0
        self.args_list = []

    def __call__(self, src, dst):
        self.called += 1
        self.args_list.append((src, dst))
        if not self._ret_values:
            return False
        return self._ret_values.pop(0)


def test_migrate_openssl_nothing_to_do(monkeypatch):
    monkeypatch.setattr(migrateopensslconf.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(migrateopensslconf, '_is_openssl_modified', IsOpensslModifiedMocked([False]))
    monkeypatch.setattr(migrateopensslconf, '_safe_mv_file', SafeMVFileMocked([False]))
    monkeypatch.setattr(os.path, 'exists', PathExistsMocked())

    migrateopensslconf.process()
    assert not os.path.exists.called
    assert not migrateopensslconf._safe_mv_file.called

    monkeypatch.setattr(migrateopensslconf, '_is_openssl_modified', IsOpensslModifiedMocked([True]))
    migrateopensslconf.process()
    assert os.path.exists.called
    assert migrateopensslconf.api.current_logger.dbgmsg
    assert not migrateopensslconf._safe_mv_file.called


def test_migrate_openssl_failed_backup(monkeypatch):
    monkeypatch.setattr(migrateopensslconf.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(migrateopensslconf, '_is_openssl_modified', IsOpensslModifiedMocked([True]))
    monkeypatch.setattr(migrateopensslconf, '_safe_mv_file', SafeMVFileMocked([False]))
    monkeypatch.setattr(os.path, 'exists', PathExistsMocked([migrateopensslconf.OPENSSL_CONF_RPMNEW]))

    migrateopensslconf.process()
    assert migrateopensslconf._safe_mv_file.called == 1
    assert migrateopensslconf._safe_mv_file.args_list[0][0] == migrateopensslconf.DEFAULT_OPENSSL_CONF
    assert migrateopensslconf.api.current_logger.errmsg


def test_migrate_openssl_ok(monkeypatch):
    monkeypatch.setattr(migrateopensslconf.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(migrateopensslconf, '_is_openssl_modified', IsOpensslModifiedMocked([True]))
    monkeypatch.setattr(migrateopensslconf, '_safe_mv_file', SafeMVFileMocked([True, True]))
    monkeypatch.setattr(os.path, 'exists', PathExistsMocked([migrateopensslconf.OPENSSL_CONF_RPMNEW]))

    migrateopensslconf.process()
    assert migrateopensslconf._safe_mv_file.called == 2
    assert migrateopensslconf._safe_mv_file.args_list[1][1] == migrateopensslconf.DEFAULT_OPENSSL_CONF
    assert not migrateopensslconf.api.current_logger.errmsg


def test_migrate_openssl_failed_migrate(monkeypatch):
    monkeypatch.setattr(migrateopensslconf.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(migrateopensslconf, '_is_openssl_modified', IsOpensslModifiedMocked([True]))
    monkeypatch.setattr(migrateopensslconf, '_safe_mv_file', SafeMVFileMocked([True, False, True]))
    monkeypatch.setattr(os.path, 'exists', PathExistsMocked([migrateopensslconf.OPENSSL_CONF_RPMNEW]))

    migrateopensslconf.process()
    assert migrateopensslconf._safe_mv_file.called == 3
    assert migrateopensslconf._safe_mv_file.args_list[2][1] == migrateopensslconf.DEFAULT_OPENSSL_CONF
    assert migrateopensslconf.api.current_logger.errmsg


def test_migrate_openssl_failed_restore(monkeypatch):
    monkeypatch.setattr(migrateopensslconf.api, 'current_logger', logger_mocked())
    monkeypatch.setattr(migrateopensslconf, '_is_openssl_modified', IsOpensslModifiedMocked([True]))
    monkeypatch.setattr(migrateopensslconf, '_safe_mv_file', SafeMVFileMocked([True]))
    monkeypatch.setattr(os.path, 'exists', PathExistsMocked([migrateopensslconf.OPENSSL_CONF_RPMNEW]))

    migrateopensslconf.process()
    assert migrateopensslconf._safe_mv_file.called == 3
    assert len(migrateopensslconf.api.current_logger.errmsg) == 2


class MockedRun(object):
    def __init__(self, raise_err):
        self.called = 0
        self.args = None
        self._raise_err = raise_err

    def __call__(self, args):
        self.called += 1
        self.args = args
        if self._raise_err:
            raise CalledProcessError(
                message='A Leapp Command Error occurred.',
                command=args,
                result={'signal': None, 'exist_code': 1, 'pid': 0, 'stdout': 'fale', 'stderr': 'fake'}
            )
        # NOTE(pstodulk) ignore return as the code in the library does not use it


@pytest.mark.parametrize('result', (True, False))
def test_is_openssl_modified(monkeypatch, result):
    monkeypatch.setattr(migrateopensslconf, 'run', MockedRun(result))
    assert migrateopensslconf._is_openssl_modified() is result
    assert migrateopensslconf.run.called == 1


@pytest.mark.parametrize('result', (True, False))
def test_safe_mv_file(monkeypatch, result):
    monkeypatch.setattr(migrateopensslconf, 'run', MockedRun(not result))
    assert migrateopensslconf._safe_mv_file('foo', 'bar') is result
    assert ['mv', 'foo', 'bar'] == migrateopensslconf.run.args



import pytest

from leapp.libraries.actor import scanrhsm
from leapp.libraries.common import rhsm
from leapp.libraries.common.testutils import CurrentActorMocked, produce_mocked
from leapp.libraries.stdlib import api
from leapp.models import RepositoryData, RepositoryFile, RHSMInfo


def mocked_get_rhsm_info(context):
    assert context, 'The actor did not provide library with valid context.'
    info = RHSMInfo()
    info.attached_skus = ['SKU1', 'SKU2']
    info.available_repos = ['Repo1', 'Repo2']
    info.enabled_repos = ['Repo2']
    info.release = '7.9'
    info.existing_product_certificates = ['Cert1', 'Cert2', 'Cert3']
    info.sca_detected = True
    return info


def test_scansubscriptionmanagementinfo(monkeypatch):
    actor_producs = produce_mocked()

    monkeypatch.setattr(rhsm, 'scan_rhsm_info', mocked_get_rhsm_info)
    monkeypatch.setattr(api, 'produce', actor_producs)
    monkeypatch.setattr(api, 'current_actor', CurrentActorMocked())

    scanrhsm.scan()

    assert actor_producs.model_instances, 'The actor did not produce any message.'
    assert len(actor_producs.model_instances) == 1, 'The actor produced more messages than expected.'

    message = actor_producs.model_instances[0]

    # The actor does not do much more than calling the `rhsm` library (which has its own tests),
    # just check that the message has not changed
    assert message.attached_skus == ['SKU1', 'SKU2']
    assert message.available_repos == ['Repo1', 'Repo2']
    assert message.enabled_repos == ['Repo2']
    assert message.release == '7.9'
    assert message.existing_product_certificates == ['Cert1', 'Cert2', 'Cert3']
    assert message.sca_detected



import base64
import io
import tarfile

from leapp.libraries.stdlib import api, CalledProcessError, run


def extract_tgz64(s):
    stream = io.BytesIO(base64.b64decode(s))
    tar = tarfile.open(fileobj=stream, mode='r:gz')
    tar.extractall('/')
    tar.close()


def enable_service(name):
    try:
        run(['systemctl', 'enable', '{}.service'.format(name)])
    except CalledProcessError:
        api.current_logger().error('Could not enable {} service'.format(name))


def write_file(name, content):
    with open(name, 'w') as f:
        f.write(content)


def ntp2chrony(root, ntp_conf, step_tickers):
    # need to skip these on pylint to avoid "function already defined" if we move to the top of file
    from leapp.libraries.actor import ntp2chrony  # pylint: disable=import-outside-toplevel

    try:
        ntp_configuration = ntp2chrony.NtpConfiguration(root, ntp_conf, step_tickers)
        ntp_configuration.write_chrony_configuration('/etc/chrony.conf', '/etc/chrony.keys',
                                                     False, True)
    except OSError as e:
        api.current_logger().error('ntp2chrony failed: {}'.format(e))
        return False, set()

    # Return ignored lines from ntp.conf, except 'disable monitor' from
    # the default ntp.conf
    return True, set(ntp_configuration.ignored_lines) - set(['disable monitor'])


def migrate_ntp(migrate_services, config_tgz64):
    # Map of ntp->chrony services and flag if using configuration
    service_map = {'ntpd': ('chronyd', True),
                   'ntpdate': ('chronyd', True),
                   'ntp-wait': ('chrony-wait', False)}

    # Minimal secure ntp.conf with no sources to migrate ntpdate only
    no_sources_directives = (
            '# This file was created to migrate ntpdate configuration to chrony\n'
            '# without ntp configuration (ntpd service was disabled)\n'
            'driftfile /var/lib/ntp/drift\n'
            'restrict default ignore nomodify notrap nopeer noquery\n')

    if not migrate_services:
        # Nothing to migrate
        return

    migrate_configs = []
    for service in migrate_services:
        if service not in service_map:
            api.current_logger().error('Unknown service {}'.format(service))
            continue
        enable_service(service_map[service][0])
        if service_map[service][1]:
            migrate_configs.append(service)

    # Unpack archive with configuration files
    extract_tgz64(config_tgz64)

    if 'ntpd' in migrate_configs:
        ntp_conf = '/etc/ntp.conf'
    else:
        ntp_conf = '/etc/ntp.conf.nosources'
        write_file(ntp_conf, no_sources_directives)

    step_tickers = '/etc/ntp/step-tickers' if 'ntpdate' in migrate_configs else ''

    conf_migrated, ignored_lines = ntp2chrony('/', ntp_conf, step_tickers)

    if conf_migrated:
        api.current_logger().info('Configuration files migrated to chrony: {}'.format(' '.join(migrate_configs)))
        if ignored_lines:
            api.current_logger().warning('Some lines in /etc/ntp.conf were ignored in migration'
                                         ' (check /etc/chrony.conf)')



from leapp.exceptions import StopActorExecutionError
from leapp.libraries.stdlib import api, CalledProcessError, run
from leapp.models import BootEntry, SourceBootLoaderConfiguration

CMD_GRUBBY_INFO_ALL = ['grubby', '--info', 'ALL']


def parse_grubby_output_line(line):
    """
    Parses a single output line of `grubby --info ALL` that has the property=value format and returns a tuple
    (property, value).

    Quotes are removed from the value.
    :param str line: A line of the grubby output.
    :returns: Tuple containing the key (boot entry property) and its value.
    :rtype: tuple
    """
    line_fragments = line.split('=', 1)
    if len(line_fragments) != 2:
        # The line does not have the property=value format, something is wrong
        raise StopActorExecutionError(
            message='Failed to parse `grubby` output.',
            details={
                'details': 'The following line does not appear to have expected format: {0}'.format(line)
            }
        )

    prop, value = line_fragments
    value = value.strip('\'"')
    return (prop, value)


def scan_boot_entries():
    """
    Scans the available boot entries.

    :rtype: list
    :returns: A list of available boot entries found in the boot loader configuration.
    """
    try:
        grubby_output = run(CMD_GRUBBY_INFO_ALL, split=True)
    except CalledProcessError as err:
        # We have failed to call `grubby` - something is probably wrong here.
        raise StopActorExecutionError(
            message='Failed to call `grubby` to list available boot entries.',
            details={
                'details': str(err),
                'stderr': err.stderr
            }
        )

    boot_entries = []
    boot_entry_data = {}
    for output_line in grubby_output['stdout']:
        if output_line == 'non linux entry':
            # Grubby does not display info about non-linux entries
            # Such an entry is not problematic from our PoV, therefore, skip it
            boot_entry_data = {}
            continue

        prop, value = parse_grubby_output_line(output_line)
        if prop == 'index':
            # Start of a new boot entry section
            if boot_entry_data:
                # There has been a valid linux entry
                boot_entries.append(
                    BootEntry(title=boot_entry_data.get('title', ''),  # In theory, the title property can be missing
                              kernel_image=boot_entry_data['kernel']))
            boot_entry_data = {}
        boot_entry_data[prop] = value

    # There was no 'index=' line after the last boot entry section, thus, its data has not been converted to a model.
    if boot_entry_data:
        boot_entries.append(BootEntry(title=boot_entry_data.get('title', ''),
                                      kernel_image=boot_entry_data['kernel']))
    return boot_entries


def scan_source_boot_loader_configuration():
    """
    Scans the boot loader configuration.

    Produces :class:`SourceBootLoaderConfiguration for other actors to act upon.
    """

    boot_loader_configuration = SourceBootLoaderConfiguration(
        entries=scan_boot_entries()
    )

    api.produce(boot_loader_configuration)



import os

from leapp.libraries.common.config import architecture
from leapp.libraries.stdlib import api
from leapp.models import CopyFile, TargetUserSpaceUpgradeTasks, UpgradeInitramfsTasks

ZFCP_CONF = '/etc/zfcp.conf'


def process():
    if not architecture.matches_architecture(architecture.ARCH_S390X):
        return
    copy_files = []
    if os.path.isfile(ZFCP_CONF):
        # the file has to be copied into the targetuserspace container first,
        # then it can be included into the initramfs ==> both messages are
        # needed to be produced
        copy_files = [CopyFile(src=ZFCP_CONF)]
        api.produce(UpgradeInitramfsTasks(include_files=[ZFCP_CONF]))
    else:
        api.current_logger().info(
            "The {} file has not been discovered. ZFCP not used."
            .format(ZFCP_CONF)
        )
    api.produce(TargetUserSpaceUpgradeTasks(copy_files=copy_files, install_rpms=['s390utils-core']))



from leapp.actors import Actor
from leapp.libraries.actor.sctplib import is_sctp_wanted
from leapp.models import ActiveKernelModulesFacts, SCTPConfig
from leapp.tags import FactsPhaseTag, IPUWorkflowTag


class SCTPConfigRead(Actor):
    """
    Determines whether or not the SCTP kernel module might be wanted.

    This actor determines whether or not the SCTP is currently used by this machine or has been quite
    recently used (1 month timeframe). In case it has been used it will issue a SCTPConfig message that
    defines the decision whether or not the SCTP module should be removed from the module blacklist on RHEL8.
    """
    name = 'sctp_read_status'
    consumes = (ActiveKernelModulesFacts,)
    produces = (SCTPConfig,)
    tags = (FactsPhaseTag, IPUWorkflowTag)

    def process(self):
        self.produce(SCTPConfig(wanted=is_sctp_wanted()))



import errno

from leapp.libraries.common.rpms import check_file_modification
from leapp.libraries.stdlib import api
from leapp.models import OpenSshConfig, OpenSshPermitRootLogin

CONFIG = '/etc/ssh/sshd_config'
DEPRECATED_DIRECTIVES = ['showpatchlevel']


def line_empty(line):
    return len(line) == 0 or line.startswith('\n') or line.startswith('#')


def parse_config(config):
    """Parse OpenSSH server configuration or the output of sshd test option."""

    # RHEL7 defaults
    ret = OpenSshConfig(
        permit_root_login=[],
        deprecated_directives=[]
    )

    in_match = None
    for line in config:
        line = line.strip()
        if line_empty(line):
            continue

        el = line.split()
        if len(el) < 2:
            continue
        value = el[1]
        if el[0].lower() == 'match':
            in_match = el[1:]
            continue

        if el[0].lower() == 'permitrootlogin':
            # convert deprecated alias
            if value == "without-password":
                value = "prohibit-password"
            v = OpenSshPermitRootLogin(value=value, in_match=in_match)
            ret.permit_root_login.append(v)

        elif el[0].lower() == 'useprivilegeseparation':
            # Record only first occurrence, which is effective
            if not ret.use_privilege_separation:
                ret.use_privilege_separation = value

        elif el[0].lower() == 'protocol':
            # Record only first occurrence, which is effective
            if not ret.protocol:
                ret.protocol = value

        elif el[0].lower() == 'ciphers':
            # Record only first occurrence, which is effective
            if not ret.ciphers:
                ret.ciphers = value

        elif el[0].lower() == 'macs':
            # Record only first occurrence, which is effective
            if not ret.macs:
                ret.macs = value

        elif el[0].lower() == 'subsystem':
            # Record only first occurrence, which is effective
            if el[1].lower() == 'sftp' and len(el) > 2 and not ret.subsystem_sftp:
                # here we need to record all remaining items as command and arguments
                ret.subsystem_sftp = ' '.join(el[2:])

        elif el[0].lower() in DEPRECATED_DIRECTIVES:
            # Filter out duplicit occurrences of the same deprecated directive
            if el[0].lower() not in ret.deprecated_directives:
                # Use the directive in the form as found in config for user convenience
                ret.deprecated_directives.append(el[0])
    return ret


def produce_config(producer, config):
    """Produce a Leapp message with all interesting OpenSSH configuration options."""

    producer(config)


def read_sshd_config():
    """Read the actual sshd configuration file."""
    try:
        with open(CONFIG, 'r') as fd:
            return fd.readlines()
    except IOError as err:
        if err.errno != errno.ENOENT:
            error = 'Failed to read sshd_config: {}'.format(str(err))
            api.current_logger().error(error)
        return []


def scan_sshd(producer):
    """Parse sshd_config configuration file to create OpenSshConfig message."""

    # direct access to configuration file
    output = read_sshd_config()
    config = parse_config(output)

    # find out whether the file was modified from the one shipped in rpm
    config.modified = check_file_modification(CONFIG)

    produce_config(producer, config)
